"""
`fields` have all possible fields where they can be used as a class variables for any Base classselfself.

Field instances will not hold the value or any date, they will be converted by Base to function descriptors,
where the data itself resides in Base objects (instances).

See `jumpscale.core.base.meta`.

Example:

```python
from enum import Enum

class Permission(Base):
    read_posts = fields.Boolean()
    write_posts = fields.Boolean()
    open_ticket = fields.Boolean()


class UserType(Enum):
    USER = "user"
    ADMIN = "admin"


class User(Base):
    emails = fields.List(fields.String())
    permissions = fields.List(fields.Object(Permission))
    custom_config = fields.Typed(dict)
    type = fields.Enum(UserType)

user = User()
emails = ["a@b.com"]
perm1 = Permission()
perm1.open_portal = True
permissions = [perm1]

user.emails = emails
user.permissions = permissions
user.save()
```

Every field type is responsible dealing with the actual value of this field in the Base objects (instances), mainly:

* validation
* converting from raw primitive types to special types if any, which would help in serialization

In addition to custom options every field can accept and define, they can be used in the following methods:

* `validate`:  raises a `ValidationError` in case it's not valid.
* `to_raw`: returns a raw (primitive type) object from a value of this field
* `from_raw`: returns a new object of the field type (if any) from a raw value

No need for `from_raw` to raise an error on e.g. type mismatch, as `validate` will do the validation.

"""
import arrow
import datetime
import ipaddress
import json
import re
import uuid

from urllib.parse import urlparse

from .factory import Factory as BaseFactory, StoredFactory


class ValidationError(Exception):
    """
    base type for any validation error
    """


class Field:
    def __init__(
        self,
        default=None,
        required=False,
        indexed=False,
        readonly=False,
        validators=None,
        stored=True,
        on_update=None,
        compute=None,
        **kwargs,
    ):
        """
        Base field for all field types, have some common options that can be used any other field type too.

        Args:
            default (any, optional): default value. Defaults to None.
            required (bool, optional): required or not. Defaults to False.
            indexed (bool, optional): indexed or not. Defaults to False.
            readonly (bool, optional): can only get the value. Defaults to False.
            validators (list of function, optional): a list of functions that takes a value and raises ValidationError if not valid. Defaults to None.
            stored (bool, optional): if the fields should be stored or not, useful for computed fields. Defaults to True.
            on_update (callable, optional): a callable that takes the instance and new value on field updates. Defaults to None.
            compute (callable, optional): a callable that takes the instance and returns a computed value for this field. `on_update` won't be called in this case. Defaults to None.
        """
        self.default = default
        self.required = required
        self.indexed = indexed
        self.readonly = readonly
        self.kwargs = kwargs

        self.validators = validators
        if self.validators is None:
            self.validators = []

        self.stored = stored
        self.on_update = on_update
        self.compute = compute

    def preprocess(self, value):
        # TODO: make from/to raw methods only for serialization
        # and let preprocess/validate do the cleanup/checking step
        # and return the correct value that current field should hold
        # so, from_raw and to_raw are used only for serialization
        return value

    def validate_with_name(self, value, name):
        """
        validates the value using validate method and prepends the field name
        in case an exception was thrown

        Args:
            value (any): in case value is not valid
            name  (str): the field name

        Raises:
            ValidationError: The original validation error with the field name prepended to the message

        """
        try:
            self.validate(value)
        except ValidationError as e:
            raise ValidationError(f"{name}: " + str(e))

    def validate(self, value):
        """
        validate value if required and call custom self.validators if any

        Args:
            value (any): in case value is not valid

        Raises:
            ValidationError: [description]
        """
        if value is None:
            if self.required:
                raise ValidationError("field is required")

        for validator in self.validators:
            validator(value)

    def from_raw(self, value):
        """
        get the value of this field from primitive raw types

        Args:
            value (any): support value by this field type (if any)

        Returns:
            any: raw value
        """
        return value

    def to_raw(self, value):
        """
        get the raw value of this field

        Args:
            value (any): current value of this field

        Returns:
            any: a primitive raw value
        """
        return value

    @property
    def trigger_updates(self):
        return callable(self.on_update)

    @property
    def computed(self):
        return callable(self.compute)


class Typed(Field):
    def __init__(self, type_, **kwargs):
        """
        Base field for any type, it will hold values of `type_`

        Args:
            type_ (type): any type (class)
            kwargs: any keyword arguments supported by `Field`
        """
        self.type = type_
        super().__init__(**kwargs)

    def validate(self, value):
        super().validate(value)
        if value is not None:
            if not isinstance(value, self.type):
                raise ValidationError(f"value '{value}' is not of type {self.type.__name__}")


class Boolean(Typed):
    def __init__(self, default=False, **kwargs):
        """
        Boolean fields to hold a `bool` value.

        values can be set using strings or numbers and will be converted to `bool` like:

        - "on", "off"
        - "yes", "no"
        - "true", "false"
        - 0, 1
        - 0, 1+2j

        Args:
            default (bool, optional): default value. Defaults to False.
            kwargs: any keyword arguments supported by `Field`
        """
        super().__init__(type_=bool, default=default, **kwargs)

    def from_raw(self, value):
        """
        get bool value from strings and numbers

        Args:
            value (str or int or float or complex)

        Returns:
            bool: boolean value
        """
        if isinstance(value, str):
            value = value.lower().strip()
            if value in ("yes", "on", "true"):
                return True
            if value in ("no", "off", "false"):
                return False
        elif isinstance(value, (int, float, complex)):
            return bool(value)

        # validate will do the check
        return value


class Integer(Typed):
    def __init__(self, default=0, min=None, max=None, **kwargs):
        """
        Intger field, the same as `Typed`, but with a type of `int`

        It can have a minimum value, if min is not set, it will ignore it.

        values can be set using strings like (will be converted to `int`):

        - "12", "1212  "

        Args:
            default (int, optional): default value. Defaults to 0.
            min (int, optional): minimum value (inclusive). Defaults to None.
            max (int, optional): maximum value (inclusive). Defaults to None.
            kwargs: any keyword arguments supported by `Field`
        """
        self.min = min
        self.max = max
        super().__init__(type_=int, default=default, min=min, **kwargs)

    def validate(self, value):
        super().validate(value)
        if value and self.min is not None:
            if value < self.min:
                raise ValidationError(f"cannot set values less than {self.min}")

        if value and self.max is not None:
            if value > self.max:
                raise ValidationError(f"cannot set values greater than {self.max}")

    def from_raw(self, value):
        if isinstance(value, str):
            try:
                value = int(value.strip())
            except ValueError:
                pass
        return value


class Float(Typed):
    def __init__(self, default=0.0, **kwargs):
        """
        Same as `Integer` field, but with a type of `float`.

        values can be set using strings like (will be converted to `float`):

        - "12.3", " 1212.23  "

        Args:
            default (float, optional): default value. Defaults to 0.0.
            kwargs: any keyword arguments supported by `Field`
        """
        super().__init__(type_=float, default=default, min=min, **kwargs)

    def from_raw(self, value):

        if isinstance(value, str):
            value = value.strip()
        try:
            value = float(value)
        except ValueError:
            pass

        return value


class String(Typed):
    def __init__(self, maxlen=None, allow_empty=True, **kwargs):
        """
        Same as `Typed`, but with a type of `str`.

        If maxlen is set, it will validate the length of the string.

        Args:
            maxlen (int): maximum length allowed. Defaults to None
            allow_empty (bool): if empty string values are allowed or not. Defaults to True
            kwargs: any keyword arguments supported by `Field`
        """
        super().__init__(type_=str, **kwargs)
        self.maxlen = maxlen
        self.allow_empty = allow_empty

    def validate(self, value):
        super().validate(value)

        if not self.allow_empty and value == "":
            raise ValidationError("field does not allow empty values")

        if self.maxlen is not None:
            if value and len(value) > self.maxlen:
                raise ValidationError(f"length of the string exceeds {self.maxlen}")


class Secret(String):
    """
    Same as `String`, but encrypted by default.

    Should be used with sensitive data.

    Args:
        kwargs: any keyword arguments supported by `String`
    """


class Object(Typed):
    def __init__(self, type_, type_kwargs=None, **kwargs):
        """
        An embedded Base object field of any type.

        Args:
            type_ (type): Base object type (class)
            type_kwargs (dict, optional): kwargs as a dict to be passed to Base instance when created. Defaults to None.
            kwargs: any keyword arguments supported by `Field`
        """
        super().__init__(type_=type_, **kwargs)
        self.type_kwargs = type_kwargs

        if self.type_kwargs is None:
            self.type_kwargs = {}

        if not self.default:
            # make it callable to create different objects
            # not a single one as a default for all
            self.default = lambda: self.type(**self.type_kwargs)

    def validate(self, value):
        """
        validate Base objects

        Args:
            value (Base): object
        """
        super().validate(value)
        if value:
            value.validate()

    def to_raw(self, obj):
        """
        get raw value of an object as dict

        Args:
            obj (Base): base object

        Returns:
            dict: raw data
        """
        if obj:
            return obj._get_data()

    def from_raw(self, data):
        """
        get an object from dict

        Args:
            data (dict): data dict

        Returns:
            Base: base object
        """
        if isinstance(data, dict):
            return self.type(**data)
        return data


class List(Field):
    def __init__(self, field, **kwargs):
        """
        A list field for any field types.

        Args:
            field (Field): a field instance of any fields, e.g. `fields.String(maxlen=14)`.
            kwargs: any keyword arguments supported by `Field`
        """
        self.field = field
        super().__init__(**kwargs)

    def validate(self, value):
        """
        validate the value of every item in the list
        Will just call the field.validate of the given field
        """
        super().validate(value)

        if value is None:
            value = []

        for item in value:
            self.field.validate(item)

    def to_raw(self, values):
        """
        get a list of values as raw

        Args:
            values (list): list of items of field type

        Returns:
            list: list of raw values
        """
        if not values:
            return []

        return [self.field.to_raw(value) for value in values]

    def from_raw(self, values):
        """
        get a list of field type from raw values

        Args:
            values (list): list of raw values

        Returns:
            list: list of objects of field type
        """
        if not values:
            return []

        return [self.field.from_raw(value) for value in values]


class Enum(Typed):
    def __init__(self, enum_type, **kwargs):
        """
        Enum field, to be used with `enum.Enum`.

        Example:

        ```python
        class UserType(Enum):
            USER = "user"
            ADMIN = "admin"

        class User(Base):
            field = fields.Enum(UserType)
        ```

        Will hold values of type `enum.Enum`.

        Args:
            enum_type (type): enum type (class)
            kwargs: any keyword arguments supported by `Field`
        """
        # default is the first value
        default = next(iter(enum_type))
        super().__init__(type_=enum_type, default=default, **kwargs)
        self.enum_type = enum_type

    def to_raw(self, enum_obj):
        """
        get enum value

        Args:
            enum_obj (enum.Enum): enum object

        Returns:
            any: enum value
        """
        if enum_obj:
            return enum_obj.value

    def from_raw(self, value):
        """
        get an enum object from value

        Args:
            value (any): any value

        Returns:
            enum: enum object of enum type of the field
        """
        try:
            return self.enum_type(value)
        except ValueError:
            # let validate() do the validation
            return value


class Email(String):
    def __init__(self, default="", **kwargs):
        """
        Email field, will validate the value of emails

        will hold string email values.

        Args:
            default (str, optional): default value. Defaults to ""
            kwargs: other keyword arguments supported by `string`
        """
        super().__init__(default=default, **kwargs)
        self.regex = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"

    def validate(self, value):
        """
        check whether provided value is a valid email representation

        Args:
            value (str)

        Raises:
            ValidationError: in case the value is not a telephone
        """
        super().validate(value)
        if value and not re.match(self.regex, value):
            raise ValidationError(f"'{value}' is not a valid Email address")


class Path(String):
    # TODO: Validate that it is working on windows
    def __init__(self, default="", **kwargs):
        """
        Path field, will validate the value of file system paths

        Will hold string path values.

        Args:
            default (str, optional): default value. Defaults to ""
            kwargs: other keyword arguments supported by `Field`
        """
        super().__init__(default, **kwargs)
        self.regex = r"^(/[^/ ]*)+/?$"

    def validate(self, value):
        """
        check whether provided value is a valid path representation

        Args:
            value (str)

        Raises:
            ValidationError: in case the value is not a telephone
        """
        super().validate(value)
        if value and not re.match(self.regex, value):
            raise ValidationError(f"'{value}' is not a valid Path")


class URL(String):
    def __init__(self, default="", **kwargs):
        """
        URL field, will validate the value of urls

        Will hold string URLs.

        Args:
            default (str, optional): default value. Defaults to ""
            kwargs: other keyword arguments supported by `Field`
        """
        super().__init__(default=default, **kwargs)
        self.regex = r"^(https?|ftp)://[^\s/$.?#].[^\s]*$"

    def validate(self, value):
        """
        check whether provided value is a valid URL representation

        Args:
            value (str)

        Raises:
            ValidationError: in case the value is not a telephone
        """
        super().validate(value)
        if value:
            url = urlparse(value)
            if not url.scheme or not url.netloc:
                raise ValidationError(f"'{value}' is not a valid URL address")


class Tel(String):
    def __init__(self, default="", **kwargs):
        """
        Telephone field, will validate the value of telephone numbers

        Will hold telephone values as strings.

        It will strip any additional characters that are not numbers.

        Args:
            default (str, optional): default value. Defaults to ""
            kwargs: other keyword arguments supported by `String`
        """
        super().__init__(default=default, **kwargs)
        self.regex = r"^\+?[0-9]{6,15}(?:x[0-9]+)?$"

    def validate(self, value):
        """
        check whether provided value is a valid telephone number representation

        Args:
            value (str)

        Raises:
            ValidationError: in case the value is not a telephone
        """
        super().validate(value)
        if value and not re.search(self.regex, value):
            raise ValidationError(f"'{value}' is not a valid Telephone")

    def from_raw(self, value):
        """clean the telephone value from unwanted signs like , - ( )"""
        if value is not None:
            value = value.replace(",", "")
            value = value.replace("-", "")
            value = value.replace("(", "")
            value = value.replace(")", "")
            value = value.replace(" ", "")
            return value
        return value


class IPMixin:
    def is_a(self, value, *types):
        """
        try creating any type of `types` from the given value

        Args:
            value (str)

        Returns:
            bool: `True` if any matched, `False` otherwise
        """
        matched = False

        for type_ in types:
            try:
                type_(value)
                matched |= True
            except ValueError:
                continue

        return matched

    def is_ipv4(self, value):
        """
        check if a given value is a v4 IP address

        Args:
            value (str): ip address

        Returns:
            bool: `True` if it's a valid v4 IP address
        """
        return self.is_a(value, ipaddress.IPv4Address)

    def is_ipv6(self, value):
        """
        check if a given value is a v6 IP address

        Args:
            value (str): IP address

        Returns:
            bool: `True` if it's a valid v6 IP address
        """
        return self.is_a(value, ipaddress.IPv6Address)

    def is_ip(self, value):
        """
        check if a given value is a v4/v6 IP address

        Args:
            value (str): IP address

        Returns:
            bool: `True` if it's a valid  v4/v6 IP address
        """
        return self.is_ipv4(value) or self.is_ipv6(value)

    def is_iface(self, value):
        """
        check if a given value is an IP interface

        Args:
            value (str): IP interface

        Returns:
            bool: `True` if it's a valid v4/v6 IP interface
        """
        return self.is_a(value, ipaddress.IPv4Interface, ipaddress.IPv6Interface)


class IPAddress(IPMixin, String):
    def __init__(self, default="0.0.0.0", **kwargs):
        """
        IP address field, will validate the value of ip address (v4 and v6)

        Will hold string ip addresses.

        Args:
            default (str, optional): default value. Defaults to ""
            kwargs: any keyword arguments supported by `Field`
        """
        super().__init__(default=default, **kwargs)

    def validate(self, value):
        """
        check whether provided value is a valid IPaddress representation
        including IPv4,IPv6 and network

        Args:
            value (str)

        Raises:
            ValidationError: in case the value is not an IPAddress
        """
        super().validate(value)
        if value and not self.is_ip(value):
            raise ValidationError(f"'{value}' is not a valid IP address")

    def from_raw(self, value):
        if isinstance(value, str):
            if value.strip().lower() == "localhost":
                value = "127.0.0.1"
        return value


class IPRange(IPMixin, String):
    def __init__(self, default="::/128", **kwargs):
        """
        ip range field, will validate the value of ip ranges (v4 and v6)

        will be stored as a string.

        Args:
            default (str, optional): default value. Defaults to ""
            kwargs: any keyword arguments supported by `Field`
        """
        super().__init__(default=default, **kwargs)

    def validate(self, value):
        """
        check whether provided value is a valid IPaddress representation
        including IPv4,IPv6 and network

        Args:
            value (str)

        Raises:
            ValidationError: in case the value is not an IPAddress
        """

        super().validate(value)
        if value and not self.is_iface(value):
            raise ValidationError(f"'{value}' is not a valid IP range/interface")


class Port(Integer):
    def __init__(self, **kwargs):
        """
        Port field, will check if the given port is within the range of 0-65535.

        Will hold integers, but also accepts string values like `"8080"`.

        Args:
            kwargs: any keyword arguments supported by `Field`
        """
        super().__init__(min=0, max=65535, **kwargs)

    def validate(self, value):
        try:
            super().validate(value)
        except ValidationError:
            raise ValidationError("value is not in range of 0-65535")


class GUID(String):
    def __init__(self, default=None, **kwargs):
        """
        UUID v4 field, will be auto-generated by default.

        Will hold the UUID as a string.

        It accepts setting UUID value by many ways and converts them to strings:

        - Strings like: `"12345678-1234-4678-9234-567812345678"`
        - Bytes like: `b"\x12\x34\x56\x78" * 4`
        - Integer like: `0x12345678123456781234567812345678`
        - UUID objects of type uuid.UUID

        Args:
            default (str, optional): default value, will be auto-generated if None. Defaults to None
            kwargs: other keyword arguments supported by `String`
        """
        if not default:
            default = lambda: str(uuid.uuid4())
        super().__init__(default=default, **kwargs)

    def validate(self, value):
        super().validate(value)

        if not value:
            return

        try:
            uuid.UUID(value, version=4)
        except ValueError as valexc:
            raise ValidationError(f"''{value}'' is invalid, {valexc}") from valexc

    def from_raw(self, value):
        """
        convert string, bytes, int to UUID object, then to a string

        Args:
            value (int or str or bytes or uuid.UUID): [description]

        Returns:
            str: UUID string
        """

        if isinstance(value, uuid.UUID):
            return str(value)

        try:
            uuid_obj = None
            if isinstance(value, int):
                uuid_obj = uuid.UUID(int=value, version=4)
            elif isinstance(value, str):
                uuid_obj = uuid.UUID(value, version=4)
            elif isinstance(value, bytes):
                uuid_obj = uuid.UUID(bytes=value, version=4)

            if uuid_obj:
                return str(uuid_obj)

            # return as is to be validated later
            return value
        except ValueError:
            # failed to get uuid from raw value
            # return as is to be validated later
            return value


class DateTimeMixin:
    def get_arrow_obj(self, value):
        """
        get an arrow object from int, float and str and `datetime.time` objects.

        Args:
            value (int or float or str): timestamp (utc) or e.g. "1998-01-03"

        Returns:
            arrow.Arrow: arrow object in utc
        """
        if isinstance(value, datetime.time):
            # convert to string, as there's no direct way i know of
            # to convert from datetime.time objects to arrow directly
            value = value.strftime(self.format)

        if isinstance(value, str):
            return arrow.Arrow.strptime(value, self.format).to("utc")

        return arrow.get(value)

    def get_timestamp(self, obj):
        """
        get a utc timestamp from datetime/date/time objects


        Args:
            obj (datetime.datetime or datetime.date or datetime.time): date/time or datetime object

        Returns:
            int or float: utc timestamp
        """
        return arrow.get(obj).to("utc").timestamp

    def from_raw(self, value):
        """
        get a datetime object from a numberic (epoch) or string value


        Args:
            value (str or int or float): value as a number or a string

        Returns:
            datetime.datetime or datetime.date or datetime.time: datetime or date/time object
        """
        if isinstance(value, (int, float, str)):
            try:
                obj = self.get_arrow_obj(value)
                if self.type == datetime.datetime:
                    return obj.datetime
                elif self.type == datetime.date:
                    return obj.date()
                elif self.type == datetime.time:
                    return obj.time()
            except (ValueError, arrow.parser.ParserError):
                # will be caught by validate
                pass
        return value

    def to_raw(self, dt_obj):
        """
        get a utc timestamp from datetime object

        Args:
            dt_obj (datetime.datetime or datetime.date or datetime.time): datetime object

        Returns:
            int or float: utc timestamp
        """
        if dt_obj:
            return self.get_arrow_obj(dt_obj).to("utc").timestamp

    def validate(self, value):
        if isinstance(self.from_raw(value), str):
            # cannot convert from string, still an invalid format
            raise ValidationError(f"'{value}' is not in the format of '{self.format}'")

        super().validate(value)


class DateTime(DateTimeMixin, Typed):
    # maybe add something like auto_now and auto_today for date/time fields
    def __init__(self, default=None, format_=None, **kwargs):
        """
        datetime field, will hold datetime.datetime objects.

        values can be set using strings in the given `format_` too like "12/1/2020" or a utc timestamp,
        they will converted to objects.

        Args:
            default (datetime): default value. Defaults to None.
            format_ (str, optional): datetime format. Defaults to "%Y-%m-%d %H:%M" if None.
            kwargs: any keyword arguments supported by `Field`
        """
        super().__init__(type_=datetime.datetime, default=default, **kwargs)
        if not format_:
            format_ = "%Y-%m-%d %H:%M"
        self.format = format_


class Date(DateTimeMixin, Typed):
    def __init__(self, default=None, format_=None, **kwargs):
        """
        date field, will hold datetime.date objects.

        values can be set using strings in the given `format_` too like "12/1/2020" or a utc timestamp,
        they will converted to objects.

        Args:
            default (date): default value. Defaults to None.
            format_ (str, optional): date format. Defaults to "%Y-%m-%d" if None.
            kwargs: any keyword arguments supported by `Field`
        """
        super().__init__(type_=datetime.date, default=default, **kwargs)
        if not format_:
            format_ = "%Y-%m-%d"
        self.format = format_


class Time(DateTimeMixin, Typed):
    def __init__(self, default=None, format_=None, **kwargs):
        """
        time field, will hold utc datetime.time objects

        values can be set using strings in the given `format_` too like "12:13" or a utc timestamp,
        they will converted to objects.

        Args:
            default (date): default value. Defaults to None.
            format_ (str, optional): time format. Defaults to "%H:%M" if None.
            kwargs: any keyword arguments supported by `Field`
        """
        super().__init__(default=default, type_=datetime.time, **kwargs)
        if not format_:
            format_ = "%H:%M"
        self.format = format_


class Bytes(Typed):
    def __init__(self, default=b"", encoding="utf-8", **kwargs):
        """
        same as string field, but will hold `bytes`.

        Args:
            default (b"", optional): default value. Defaults to b""
            encoding: encoding to be used when serializing the value. Defaults to "utf-8"
            kwargs: any keyword arguments supported by `Field`
        """
        self.encoding = encoding
        super().__init__(default=default, type_=bytes, **kwargs)

    def from_raw(self, value):
        if isinstance(value, str):
            return value.encode(self.encoding)
        return value

    def to_raw(self, value):
        if value:
            return value.decode(self.encoding)


class Json(String):
    def __init__(self, default="{}", encoding="utf-8", **kwargs):
        """
        Json field, will check if the value is a valid json string.

        It will hold json strings, if the value is not string, it will be converted.

        Args:
            default (str, optional): default value. Defaults to "{}"
            encoding: encoding to be used when serializing the value. Defaults to "utf-8"
            kwargs: other keyword arguments supported by `String`
        """
        self.encoding = encoding
        super().__init__(default=default, **kwargs)

    def validate(self, value):
        """
        check whether provided value is a valid json

        Args:
            value (str)

        Raises:
            ValidationError: in case the value isn't a valid json
        """
        super().validate(value)

        if not value:
            return

        # if it's a string, try to load it
        try:
            json.loads(value, encoding=self.encoding)
        except Exception as e:
            raise ValidationError(f"'{value}' isn't a valid json, {e}") from e

    def from_raw(self, value):
        """
        convert non-string values to json string

        Args:
            value (any): non-string value

        Returns:
            str: a json string
        """
        if not isinstance(value, (str, bytes, bytearray)):
            try:
                value = json.dumps(value)
            except:
                pass
        return value


class Factory(Field):
    def __init__(self, type_, factory_type=None, **kwargs):
        """
        A factory field for any `Base` type, also, you can specify your factory type/class

        Example:

        ```python
        class User(Base):
            name = fields.String()

        class Server(Base):
            users = fields.Factory(User)
        ```

        Another example with a custom factory class:

        ```python
        class User(Base):
            name = fields.String()

        class UserFactory(StoredFactory):

            def list_from_remote(self):
                # list users from remote storage
                # ...

        class Server(Base):
            users = fields.Factory(User, factory_type=UserFactory)
        ```

        Args:
            type_ (Base): any base type to be used by the factory
            factory_type (`BaseFactory`, optional): factory class/type. Defaults to None.
        """
        # value type will be factory
        # make sure stored default value is True
        kwargs["stored"] = kwargs.get("stored", True)
        super().__init__(readonly=True, **kwargs)
        # but we keep the type of any Base class
        # so, we can init a Factory with it
        self.type = type_
        if factory_type:
            self.factory_type = factory_type
        else:
            if self.stored:
                self.factory_type = StoredFactory
            else:
                self.factory_type = BaseFactory

    def validate(self, value):
        super().validate(value)
        if not isinstance(value, self.factory_type):
            raise ValidationError(f"factory type is not {self.factory_type}")

    def from_raw(self, value):
        return value

    def to_raw(self, value):
        return None



"""idgenerator module helps with generating ids, guids, integers, chars, passwords, capnp id, a choice of a sequence

```
JS-NG> j.data.idgenerator.guid()                                                                    
'c1d14970-f17f-49a3-aa85-f722013ee448'

JS-NG> j.data.idgenerator.password(5)                                                               
'b6~Sl'

JS-NG> j.data.idgenerator.capnp_id()                                                                
'0xa414b890b73d0940'

JS-NG> j.data.idgenerator.random_int()                                                              
7

JS-NG> j.data.idgenerator.random_int(0, 5)                                                          
2
```
"""
import random
import string
import uuid
from typing import List


def random_int(from_: int = 0, to: int = 10) -> int:
    """Generate random int within range from_, to

    Arguments:
        from_ {int} -- lower limit, default 0
        to {int} -- upper limit

    Returns:
        int -- random number
    """
    return random.randint(from_, to)


def incrementor_id():
    # user redis.incr.
    raise NotImplementedError()


def guid() -> str:
    """Gets a uuid4

    Returns:
        str -- uuid4
    """
    return str(uuid.uuid4())


def nfromchoices(n: int, choices: List[str]) -> str:
    """Gets string from choices list

    Arguments:
        n {int} -- how many
        choices {List[str]} -- choices to pick from

    Returns:
        str -- joined result of n choices.
    """
    return "".join([random.choice(choices) for _ in range(n)])


def chars(nchars: int) -> str:
    """Gets a random string with length of n

    Arguments:
        nchars {int} -- [description]

    Returns:
        str -- [description]
    """
    choices = string.ascii_letters + string.digits
    return nfromchoices(nchars, choices)


def nbytes(nbytes: int) -> bytearray:
    """return bytearray of length n

    Arguments:
        nbytes {int} -- number of bytes to generate

    Returns:
        bytearray -- result
    """
    out = bytearray()
    for n in range(nbytes):
        out.append(random_int(0, 255))
    return out


def password(nchars: int) -> str:
    """Return a password of length nchars

    Arguments:
        nchars {int} -- [description]

    Returns:
        str -- password
    """
    choices = string.printable
    return nfromchoices(nchars, choices)


def capnp_id() -> str:
    """
    Generates a valid id for a capnp schema.

    Returns:
        str -- capnp id
    """
    # the bitwise is for validating the id check capnp/parser.c++
    return hex(random.randint(0, 2 ** 64) | 1 << 63)



import tarfile


def istar(path):
    """check if the file is .tar format
    
    Arguments:
        path (str) : the path for the file
    """
    return tarfile.is_tarfile(path)


def compress(source, output):
    """make an archive file from directory or file
    
    Arguments:
        source (str) : the path for the file or the directory
        output (str) : the path for the output
    """
    with tarfile.open(output, "w") as output:
        output.add(source)


class Reader:
    """handle the reading operation on tar file
    
    Arguments:
        path (str) : the path for tar file
    """

    def __init__(self, path):

        self.path = path
        self.file = tarfile.TarFile.open(self.path, "r")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.file.close()

    def get_content(self):
        """get list of the content in tar file"""
        return self.file.getnames()

    def extract(self, output):
        """extract all the files from the archive to a directory.
        Args:
            output (str) : the path for the output folder
        """
        self.file.extractall(path=output)



import re
from jumpscale.loader import j


def pascalify_name(name):
    return "".join([x.capitalize() for x in name.replace("_", "-").split("-")])


class Property:
    def __init__(self):
        self.index = False
        self.index_key = False
        self.index_text = False
        self.unique = False
        self.type = None
        self.comment = ""
        self.defaultvalue = ""
        self.name = ""
        self.pointer_type = ""
        self.prop_type = ""

    @classmethod
    def get_id_prop(cls):
        prop = cls()
        prop.prop_type = "I"
        prop.name = "id"
        prop.type = j.data.types.get_js_type(prop.prop_type)
        return prop

    def __repr__(self):
        return str(self.__dict__)

    __str__ = __repr__

    @property
    def url_to_class_name(self):
        url = self.defaultvalue
        urlparts = url.split(".")
        return "".join(x.capitalize() for x in urlparts)


class Schema:
    def __init__(self):
        self.url = ""
        self.system_props = {}
        self.props = {"id": Property.get_id_prop()}

    @property
    def url_to_class_name(self):
        return "".join(x.capitalize() for x in self.url.split("."))

    def get_enums_required(self):
        enums = []
        for prop_name, prop in self.props.items():
            if prop.prop_type == "E":
                cleanname = pascalify_name(prop.name)
                enums.append({"name": cleanname, "vals": [pascalify_name(x) for x in prop.defaultvalue.split(",")]})
        return enums

    def get_classes_required(self):
        classes = []
        for prop_name, prop in self.props.items():
            if prop.prop_type in ["O", "LO"] and "." in prop.defaultvalue:
                classes.append(prop.url_to_class_name)
        return classes

    def get_dependencies(self):

        return {enum: self.get_enums_required(), classes: self.get_classes_requireds()}


def _normalize_string(text):
    """Trims the text, removes all the spaces from text and replaces every sequence of lines with a single line.

    Args:
        text (str): The schema definition.

    Returns:
        str: The normalized text.
    """
    text = text.strip()
    result = ""
    whitespaces = [" ", "\t"]
    in_comment = False
    in_string = False
    string_char = None
    for c in text:
        if c == "#":
            in_comment = True
        if c == "\n":
            in_comment = False
        if c == '"' or c == "'":
            if in_string and c == string_char:
                in_string = False
            elif not in_string and not in_comment:
                in_string = True
                string_char = c

        if not in_comment and not in_string and c in whitespaces:
            continue
        if c == "\n" and result.endswith("\n"):
            continue
        result += c
    return result


def _correct_url_number(text):
    """Checks that the schema definition contains only one url definition.

    Args:
        text (str): The schema definition.

    Returns:
        bool: Whether only one url is defined.
    """
    url_count = text.count("@url")
    return url_count == 1


def _parse_system_prop(line):
    """Returns the name and value of a system property (The line is preceded by @).

    Args:
        line (str): The definition of the system property.

    Returns:
        str, str: Pair of name, value.
    """
    return line[1:].split("#")[0].split("=")


def _name_in_correct_form(name):
    """Checks whether the name is a sequence of alphanumberic characters and _ and starts with _.

    Args:
        name (str): The name of the property.

    Returns:
        bool: True if the name is well formed. False otherwise.
    """
    return name[0] == "_" or name[0].isalpha() and name.replace("_", "").isalnum()


def _is_float(value):
    """Checks if the value is float. Which means it contains two non empty integer parts separated by .

    Args:
        value (str): The value.

    Returns:
        bool: True if the value represents a float.
    """
    if "." not in value:
        return False
    a, b = value.split(".", 1)
    return a.isdigit() and b.isdigit()


def _infer_type(value):
    """Infers the type from the value.
    one of:
    1. int: contains digits only.
    2. float: two nonempty parts of integers separated by .
    3. string: nonempty sequence of characters inside " or '
    4. bool: literals true or false.
    5. list: equals surrounded by [].

    Args:
        value (str): The value which the type is infered from.

    Raises:
        RuntimeError: If no types can be infered

    Returns:
        str: The type of the value.
    """
    if len(value) >= 2 and (value[0] == '"' and value[-1] == '"' or value[0] == "'" and value[-1] == "'"):
        return "S"
    elif _is_float(value):
        return "F"
    elif value.isdigit():
        return "I"
    elif value == "[]":
        return "L"
    elif value == "true" or value == "false":
        return "B"
    else:
        raise RuntimeError(f"Can't infer the type of {value}")


def _parse_prop(line):
    """Parses a line which defines a property.

    Args:
        line (str): The property description.

    Raises:
        RuntimeError: If the description is malformed.

    Returns:
        str, Property: Name of the property and Property object defining the property.
    """
    RE = r"(&?)([^=*]+)((?:\*|\*\*|\*\*\*)?)\=([^(!#]*)((?:\(.*?\))?)((?:\![^#]*)?)((?:#.*)?)"
    match = re.match(RE, line)
    if match is None:
        raise RuntimeError("Can't parse schema")
    prop = Property()
    unique, name, qualifier, default_value, prop_type, pointer_type, comment = match.groups()
    if name == "id":
        raise NameError("id is reserved and can't be a name of property.")
    pointer_type = pointer_type[1:] if pointer_type else pointer_type
    prop.pointer_type = pointer_type
    prop_type = prop_type[1:-1] if prop_type else prop_type

    if not prop_type:
        prop_type = "S"
    prop.prop_type = prop_type

    comment = comment[1:] if comment else comment
    prop.name = name
    prop.unique = unique == "&"
    prop.index = name == "name" or prop.unique or qualifier == "*"
    prop.index_key = qualifier == "**"
    prop.index_text = qualifier == "***"
    prop_type = prop_type or _infer_type(default_value)
    if len(default_value) >= 2 and (
        default_value[0] == "'" and default_value[-1] == "'" or default_value[0] == '"' and default_value[-1] == '"'
    ):
        default_value = default_value[1:-1]
    prop.defaultvalue = pointer_type or default_value
    prop.type = j.data.types.get_js_type(prop_type, prop.defaultvalue)
    prop.comment = comment
    return prop.name, prop


def parse_schema(text):
    """Parses a schema from string.

    Args:
        text (str): The schema definition.

    Raises:
        RuntimeError: Thrown when the schema can't be parsed.

    Returns:
        Schema: Schema object representing the schema.
    """
    # print(f"parsing \n {text}")
    text = _normalize_string(text)
    if text.count("@url") != 1:
        raise RuntimeError("The number of urls must be equal to one.")
    lines = text.split("\n")
    schema = Schema()
    for line in lines:
        if line.startswith("@"):
            name, value = _parse_system_prop(line)
            schema.system_props[name] = value
        elif line.startswith("#") or not line.strip():
            continue
        else:
            name, prop = _parse_prop(line)
            schema.props[name] = prop
    schema.url = schema.system_props["url"]
    return schema



"""`jsync` is syncing tool to sync over certain set of directories against remote machine.
## list available clients
```
~> poetry run jsync list-ssh-clients
['xmonader']
```

## sync with a client certain set of paths
```
~> poetry run jsync sync --clients "xmonader" --paths "~/wspace/tq,/tmp/proj:/tmp/proj2"
['xmonader'] {'~/wspace/tq': '~/wspace/tq', '/tmp/proj': '/tmp/proj2'}
```
"""

import click
from jumpscale.loader import j


@click.command()
def list_ssh_clients():
    # TODO: show terminal table with host information
    return print(j.clients.sshclient.list_all())


# jsync --clients "xmonader,client2" --paths "/home/xmonader/wspace/tq:/tmp/tq,..."
@click.command()
@click.option("--clients")
@click.option("--paths")
@click.option("--nosync", is_flag=True, default=False, type=bool)
def sync(clients, paths, nosync=False):
    clients = [cl_name.strip() for cl_name in clients.split(",")]
    paths_dict = {}
    for watched_path_info in paths.split(","):
        parts = watched_path_info.split(":")
        if len(parts) == 1:
            src, dest = parts[0], parts[0]
        else:
            src, dest = parts[0], parts[1]
        src = j.sals.fs.expanduser(src)
        paths_dict[src] = dest

    j.logger.info("clients: {}, paths {} ".format(clients, paths_dict))
    syncer = j.tools.syncer.Syncer(clients, paths_dict)
    syncer.start(sync=not nosync)


@click.group()
def cli():
    pass


cli.add_command(sync)
cli.add_command(list_ssh_clients)

if __name__ == "__main__":
    cli()



"""Generator based on JS-NG fields. takes in parsed schemas and generates Python classes based on JS-NG fields.

"""

from .plugin import Plugin

types_map = {
    "": "String",
    "S": "String",
    "O": "Object",
    "I": "Integer",
    "F": "Float",
    "B": "Boolean",
    "L": "List",
    "LS": "List",
    "LI": "List",
    "LF": "List",
    "LO": "List",
    "E": "Enum",
    "D": "DateTime",
    "T": "Date",
    "guid": "GUID",
    "email": "Email",
    "dict": "Object",
    "ipaddr": "IPAddress",
    "ipaddress": "IPAddress",
    "iprange": "IPRange",
    "json": "Json",
    "bytes": "Bytes",
    "I64": "Float",
}


def get_prop_line(prop):
    prop_type = prop.prop_type
    if prop_type == "B":
        if prop.defaultvalue.lower() == "true":
            prop.defaultvalue = "True"
        elif prop.defaultvalue.lower() == "false":
            prop.defaultvalue = "False"

    python_type = types_map.get(prop_type)
    line = f"{prop.name} = "
    # print(f"\n\n{prop.name} => {prop} \n\n")
    # primitive with a default or not.
    if prop_type == "E":
        line += f"fields.{python_type}({prop.name.capitalize()})"
    elif prop_type == "O":
        line += f"fields.{python_type}({prop.url_to_class_name})"
    elif prop_type == "LO" and prop.defaultvalue and prop.defaultvalue != "[]":
        line += f"fields.List(fields.Object({prop.url_to_class_name}))"
    elif prop_type == "LO" and not prop.defaultvalue:
        line += "fields.List(fields.Object(Base))"
    elif python_type == "L" and not prop.defaultvalue:
        line += " fields.List(fields.Object())"
    elif python_type == "L" and prop.defaultvalue:
        line += f" fields.List(fields.Object({prop.defaultvalue}))"
    elif len(prop_type) > 1 and prop_type[0] == "L":
        line += f"fields.List(fields.{types_map[prop_type[1:]]}())"
    elif prop_type in ["I", "F", "B"] and not prop.defaultvalue:
        line += f"fields.{python_type}()"
    elif prop_type in ["I", "F", "B"] and prop.defaultvalue:
        line += f"fields.{python_type}(default={prop.defaultvalue})"
    elif prop_type == "S":
        line += f'fields.String(default="{prop.defaultvalue}")'
    elif prop_type in ["T", "D"]:
        line += f"fields.{types_map[prop_type]}()"
    elif prop_type == "dict":
        line += "fields.Typed(dict)"
    elif prop_type in ["ipaddress", "ipaddr", "iprange"] and prop.defaultvalue:
        line += f'fields.{python_type}(default="{prop.defaultvalue}")'
    else:
        line += f"fields.{python_type}()"

    return line


SINGLE_TEMPLATE = """

class {{generated_class_name}}(Base):
{%- for prop in generated_properties.values() %}
    {{get_prop_line(prop)}}
{%- endfor %}

"""

TEMPLATE = """
#GENERATED CLASS DONT EDIT
from jumpscale.core.base import Base, fields
from enum import Enum

{%- for enum in enums %}

class {{enum['name']}}(Enum):
    {%- for enumval in enum['vals'] %}
    {{enumval}} = {{loop.index0}}
    {%- endfor %}
{%- endfor %}

{{classes_generated}}

"""


class JSNGGenerator(Plugin):
    def __init__(self):
        super().__init__()

        self._single_template = SINGLE_TEMPLATE
        self._template = TEMPLATE
        self._get_prop_line = get_prop_line
        self._types_map = types_map



import click

from jumpscale.core.config import get_default_config, update_config, get_config
import toml


def decode_toml_value(s):
    try:
        return toml.TomlDecoder().load_value(s)[0]
    except Exception:
        return s


def encode_toml_value(s):
    return toml.TomlEncoder().dump_value(s)


def format_config_parameter(name, value):
    if isinstance(value, dict):
        print_dict(value, f"{name}.")
    else:
        return "%s = %s" % (name, encode_toml_value(value))


def print_dict(data, path):
    for name, value in data.items():
        if isinstance(value, dict):
            print_dict(value, f"{path}{name}.")
        else:
            encoded = encode_toml_value(value)
            print(f"{path}{name} = {encoded}")


def validate_type(a, b):
    typea = type(a).__name__
    typeb = type(b).__name__
    if typea != typeb:
        raise TypeError(f"Type mismatch: Expected {typea} found {typeb}.")


def format_config(config):
    print_dict(config, "")


def traverse_config(name):
    path = name.split(".")
    config = get_config()
    prop = config
    for i, p in enumerate(path[:-1]):
        prop = prop[p]
    return config, prop, path[-1]


@click.group()
def config():
    pass


@config.command()
def defaults():
    """list default configuration."""
    format_config(get_default_config())


@config.command()
def list_all():
    """list all configurations."""
    format_config(get_config())


@config.command()
@click.argument("name")
def get(name):
    """get the key from the configuration, use dots to access nested values."""
    try:
        _, obj, prop = traverse_config(name)
        if not isinstance(obj, dict):
            raise KeyError()
        value = obj[prop]
        click.echo(format_config_parameter(name, value))
    except KeyError:
        click.echo('Key doens\'t exist. Type "jsctl config list-all" to see all valid configurations.')


@config.command()
@click.argument("name")
@click.argument("value")
def update(name, value):
    """set the passed key by the corresponding toml value."""
    try:
        config, obj, prop = traverse_config(name)
        if not isinstance(obj, dict) or prop not in obj:
            raise KeyError()
        value = decode_toml_value(value)
        validate_type(obj[prop], value)
        obj[prop] = value
        update_config(config)
    except KeyError:
        click.echo('Key doens\'t exist. Type "jsctl config list-all" to see all valid configurations.')
    except TypeError as e:
        click.echo(str(e))
    else:
        click.echo("Updated.")


@click.group()
def cli():
    pass


cli.add_command(config)

if __name__ == "__main__":
    cli()



"""This module does all the work for serialization/deserialization around pickle, base64, json, msgpack, pickle, dill, toml
```
JS-NG> obj = {"name":"username", "list":[1,3,4,7], "n":5}                                           

JS-NG> j.data.serializers.json.dumps(obj)                                                           
'{"name": "username", "list": [1, 3, 4, 7], "n": 5}'

JS-NG> j.data.serializers.toml.dumps(obj)                                                           
'name = "username"\nlist = [1, 3, 4, 7]\nn = 5\n'

JS-NG> j.data.serializers.yaml.dumps(obj)                                                           
'list:\n- 1\n- 3\n- 4\n- 7\nn: 5\nname: username\n'

JS-NG> j.data.serializers.msgpack.dumps(obj)                                                        
b'\x83\xa4name\xa8username\xa4list\x94\x01\x03\x04\x07\xa1n\x05'
```
"""

from . import base64
from . import json
from . import lzma
from . import msgpack
from . import pickle
from . import toml
from . import yaml
from . import dill



import json


def dumps(obj):
    """dump dict object into json stream

    Arguments:
        obj (dict) : the dict which will be dumped

    Returns:
        string : the json stream
    """
    return json.dumps(obj)


def dump_to_file(file_path, obj):
    """Writes the dumped obj to a file

   Args:
       file_path (str): path to write to
       obj (dict): the dict which will be dumped
   """
    with open(file_path, "w") as fp:
        json.dump(obj, fp)


def loads(s):
    """loads the data from json string into dict

    Arguments:
        s (string) : the json stream

    Returns:
        dict : the loaded data from json stram
    """
    if isinstance(s, bytes):
        s = s.decode("utf-8")
    return json.loads(s)


def load_from_file(file_path):
    """Loads data from file to a dict

    Args:
        file_path (str): path of the json file
    """
    with open(file_path) as fp:
        obj = json.load(fp)
    return obj



import pytoml


def dumps(d):
    """dump dict object into toml stream 
    
    Arguments:
        d (dict) : the dict which will be dumped     
    
    Returns:
        string : the toml stream
    """
    s = pytoml.dumps(d)
    return s


def loads(s):
    """loads the data from toml string into dict
    
    Arguments:
        s (string) : the toml stream
    
    Returns:
        dict : the loaded data from toml stram
    """
    d = pytoml.loads(s)
    return d




from gevent import monkey

monkey.patch_all(subprocess=False)  # noqa: E402
import click
import os
import pathlib
import sys

from jumpscale.loader import j  # noqa:
from jumpscale.core.config.config import get_config

BASE_CONFIG_DIR = os.path.join(os.environ.get("HOME", "/root"), ".jsng")
HISTORY_FILENAME = os.path.join(BASE_CONFIG_DIR, "history.txt")


@click.command()
@click.argument("command", required=False)
def run(command):
    """Executes the passed command and initiates a jsng shell if no command is passed."""
    os.makedirs(BASE_CONFIG_DIR, exist_ok=True)
    pathlib.Path(HISTORY_FILENAME).touch()
    if command is None:
        config = get_config()
        if config["shell"] == "ipython":
            from IPython import embed
            sys.exit(embed(colors="neutral"))
        else:
            from jumpscale.shell import ptconfig
            from ptpython.repl import embed
            sys.exit(embed(globals(), locals(), configure=ptconfig, history_filename=HISTORY_FILENAME))
    else:
        sys.exit(print(eval(command)))



"""This module helps with the common operations of jinja2 and reduces the boilerplate around it


## Getting environment

```python
from jinja2 import (
    Environment,
    FileSystemLoader,
    select_autoescape,
    StrictUndefined,
    Template,
)
env = Environment(loader=FileSystemLoader(templates_path), autoescape=select_autoescape(["html", "xml"]),)

```
but you can easily get the Environment without worrying too much about the syntax with `j.tools.jinja2.get_env()` and that's it.


## Getting a template from path or text
and same for getting a specific template object from a text or a file, but you can easily do `get_template(template_path=...)` or `get_template(template_text=...)`


## Rendering a template with data
you can render from a file path or a text directly using `j.tools.jinja2.render_template` and pass `template_text` in case of a string or `template_path` in case of a file path.

e.g

```python
        data = dict(
            generated_class_name=schema.url_to_class_name,
            generated_properties=schema.props,
            types_map=self._types_map,
            enums=schema.get_enums_required(),
            classes=schema.get_classes_required(),
            get_prop_line=self._get_prop_line,
        )
        return j.tools.jinja2.render_template(template_text=self._single_template, **data)
```
"""

from jumpscale.loader import j
from jinja2 import (
    Environment,
    FileSystemLoader,
    select_autoescape,
    StrictUndefined,
    Template,
)


def get_env(templates_path):
    """get an environment from templates root path

    Args:
        templates_path (str): root path of all templates

    Returns:
        jinja2.Environment: Jinja2 env
    """
    return Environment(loader=FileSystemLoader(templates_path), autoescape=select_autoescape(["html", "xml"]),)


def get_template(template_path=None, template_text=None):
    """returns jinja2 template

    Args:
        template_path (str, optional): location of the template. Defaults to None.
        template_text (str, optional): text of the template. Defaults to None.

    Raises:
        j.exceptions.Input: If both template_path and template_text are specified
        j.exceptions.Input: If no input was specified

    Returns:
        jinja2.Template: Jinja2 template
    """

    if template_path and template_text:
        raise j.exceptions.Input("Can only specify template_path or template_text")
    if not template_path and not template_text:
        raise j.exceptions.Input("Need to specify either template_path or template_text")

    if template_path:
        template_text = j.sals.fs.read_file(template_path)

    return Template(template_text, undefined=StrictUndefined)


def render_template(template_path=None, template_text=None, dest=None, **kwargs):
    """load the template if dest is specified will write in specified path, renders with specified kwargs

    Args:
        template_path (str, optional): location of the template. Defaults to None.
        template_text (str, optional): text of the template. Defaults to None.
        dest (str, optional): path to write rendered template in. Defaults to None.

    Raises:
        j.exceptions.Base: If rendering failed

    Returns:
        str: Rendered template
    """
    template = get_template(template_path=template_path, template_text=template_text)

    try:
        template_string = template.render(**kwargs)
    except Exception as e:
        raise j.exceptions.Base(e)

    if dest:
        j.sals.fs.mkdirs(j.sals.fs.dirname(dest))
        j.sals.fs.write_file(dest, template_string)
    return template_string


def render_code_python(
    obj_key=None, template_path=None, template_text=None, dest=None, objForHash=None, name=None, **kwargs,
):
    # TODO
    pass



import git
from jumpscale.clients.base import Client
from jumpscale.core.base import fields
from jumpscale.loader import j


class GitClient(Client):
    path = fields.String()

    def __init__(self):
        super().__init__()
        self.__repo = None

    @property
    def repo(self):
        if not self.__repo:
            self.__repo = git.Repo(self.path)
        return self.__repo

    def set_remote_url(self, url, remote_name="origin"):
        remote = self.repo.remote(remote_name)
        remote.set_url(url)

    @property
    def remote_url(self):
        return self.repo.remote().url

    @property
    def branch_name(self):
        return self.repo.active_branch.name

    def get_modified_files(self):
        """returns local changes in the repo

        Returns:
            dict: dict containing different types of changes(check git status man)
        """
        modified_files = self.repo.git.status(porcelain=True).splitlines()
        result_format = {}
        for mod_file in modified_files:
            if "??" in mod_file:
                continue
            state, file_name = mod_file.split()
            result_format.setdefault(state, [])
            result_format[state].append(file_name)
        untracked_files = self.repo.untracked_files
        if untracked_files:
            result_format["N"] = untracked_files
        return result_format

    def pull(self):
        """Pulls from origin

        Raises:
            j.exceptions.Input: if there is locaal changes
        """
        if self.get_modified_files():
            raise j.exceptions.Input(message="Cannot pull:{}, files waiting to commit".format(self.path))
        self.repo.git.pull()

    def commit(self, message, add_all=True):
        """adds a commit

        Args:
            message (str): commit message
            add_all (bool, optional): will add all changes before commiting. Defaults to True.

        Returns:
            [type]: [description]
        """
        if add_all and self.get_modified_files():
            self.repo.git.add("-A")
        if self.repo.index.diff("HEAD"):
            return self.repo.index.commit(message)



import pylzma

def compress(obj):
    """compress string with lzma algorithm
    
    Arguments:
        obj (string) : the string will be encoded
    
    Returns:
        bytes : the compressed bytes
    """
    return pylzma.compress(obj)

def decompress(s):
    """decompress lzma bytes to original obj
    
    Arguments:
        s (bytes) : the bytes will be compressed
    
    Returns:
        (string) : the decompressed string
    """
    return pylzma.decompress(s)


def export_module_as():

    from jumpscale.loader import j
    from .logging import Logger, RedisLogHandler, LogHandler

    logger = Logger()
    config = j.core.config.config.get_config().get("logging")

    if config:
        # configure redis handler if enabled
        if config["redis"]["enabled"] and j.core.db:
            redis_config = config["redis"]
            redis_handler = RedisLogHandler(
                max_size=redis_config["max_size"], dump=redis_config["dump"], dump_dir=redis_config["dump_dir"]
            )
            logger.add_custom_handler("redis", redis_handler, serialize=True, level=redis_config["level"])

        # configure filesystem handler if enabled
        if config["filesystem"]["enabled"]:
            fs_config = config["filesystem"]
            logger.add_handler(sink=fs_config["log_dir"], level=fs_config["level"], rotation=fs_config["rotation"])

    return logger



import math
import loguru
import msgpack
import json

from abc import ABC, abstractmethod

from jumpscale.loader import j


LEVELS = {
    10: "DEBUG",
    20: "INFO",
    30: "WARNING",
    40: "ERROR",
    50: "CRITICAL",
}


class LogHandler(ABC):
    """the interface every cutom log handler should implement"""

    @abstractmethod
    def _handle(self, message, **kwargs):
        pass


class Logger:
    def __init__(self, appname: str = "init"):
        """Logger init method
        """
        self._logger = loguru.logger.bind(appname=appname)
        self._appname = appname

    @property
    def appname(self):
        return self._appname

    def set_appname(self, appname):
        self._appname = appname
        self._logger = self._logger.bind(appname=appname)

    def add_handler(self, *args, **kwargs):
        """Add handler to the logger
        takes the same parameters of loguru.logger.add
        """
        self._logger.add(*args, **kwargs)

    def add_custom_handler(self, name: str, handler: LogHandler, *args, **kwargs):
        """Add custom log handler

        Arguments:
            handler {LogHandler} -- handler function
        """
        setattr(self, name, handler)
        self._logger.add(handler._handle, **kwargs)

    def _log(self, level, message, *args, category, data, exception=None):
        self._logger.opt(depth=2, exception=exception).bind(category=category, data=data).log(level, message, *args)

    def debug(self, message, *args, category: str = "", data: dict = None):
        """Log debug message
        """
        self._log("DEBUG", message, *args, category=category, data=data)

    def info(self, message, *args, category: str = "", data: dict = None):
        """Log info message
        """
        self._log("INFO", message, *args, category=category, data=data)

    def warning(self, message, *args, category: str = "", data: dict = None):
        """Log warning message
        """
        self._log("WARNING", message, *args, category=category, data=data)

    def error(self, message, *args, category: str = "", data: dict = None):
        """Log error message
        """
        self._log("ERROR", message, *args, category=category, data=data)

    def critical(self, message, *args, category: str = "", data: dict = None):
        """Log critical message
        """
        self._log("CRITICAL", message, *args, category=category, data=data)

    def exception(
        self, message, *args, category: str = "", data: dict = None, level: int = 40, exception: Exception = None
    ):
        """Log exception message
        """
        self._log(LEVELS.get(level, 40), message, *args, category=category, data=data, exception=exception)


class RedisLogHandler(LogHandler):
    def __init__(self, max_size: int = 1000, dump: bool = True, dump_dir: str = None):
        self._max_size = max_size
        self._dump = dump
        self._dump_dir = dump_dir or "/tmp"
        self._rkey = "logs:%s:records"
        self._rkey_incr = "logs:%s:incr"
        self.__db = None

        if self._dump_dir:
            j.sals.fs.mkdirs(self._dump_dir)

    @property
    def _appname(self):
        return j.core.application.appname

    @property
    def _db(self):
        if self.__db is None:
            self.__db = j.core.db
        return self.__db

    @property
    def max_size(self):
        return self._max_size

    @property
    def dump(self):
        return self._dump

    @property
    def dump_dir(self):
        return self._dump_dir

    def _map_identifier(self, identifier):
        part = math.ceil(identifier / self.max_size) * self.max_size
        index = (identifier % self.max_size) - 1
        return part, index

    def _dump_records(self, appname, path):
        j.sals.fs.mkdir(j.sals.fs.parent(path))
        records = self._db.lrange(self._rkey % appname, 0, self.max_size - 1)
        j.sals.fs.write_bytes(path, msgpack.dumps(records))

    def _process_message(self, message):
        record = json.loads(message)["record"]
        appname = record["extra"]["appname"]
        record_id = self._db.incr(self._rkey_incr % appname)
        return dict(
            id=record_id,
            appname=appname,
            message=record["message"],
            level=record["level"]["no"],
            linenr=record["line"],
            file=record["file"],
            processid=record["process"]["id"],
            context=record["function"],
            epoch=record["time"]["timestamp"],
            exception=record["exception"],
            category=record["extra"].get("category", ""),
            data=record["extra"].get("data", {}),
        )

    def _clean_up(self, appname):
        self._db.ltrim(self._rkey % appname, self.max_size, -1)

    def _handle(self, message: str, **kwargs):
        """Logging handler

        Arguments:
            message {str} -- message string
        """
        if not self._db.is_running():
            return

        record = self._process_message(message)
        appname = record["appname"]

        rkey = self._rkey % appname
        self._db.rpush(rkey, json.dumps(record))

        if self._db.llen(rkey) > self.max_size:
            if self.dump:
                part, _ = self._map_identifier(record["id"] - 1)
                path = j.sals.fs.join_paths(self.dump_dir, appname, "%s.msgpack" % part)
                self._dump_records(appname, path)

            self._clean_up(appname)

    def records_count(self, appname: str = "init") -> int:
        """Gets total number of the records of the app

        Arguments:
            appname {str} -- app name

        Returns:
            init -- total number of the records
        """
        count = self._db.get(self._rkey_incr % appname)
        if count:
            return int(count)
        return 0

    def record_get(self, identifier: int, appname: str = "init") -> dict:
        """Get app log record by its identifier

        Arguments:
            identifier {int} -- record identifier
            appname {str} -- app name

        Returns:
            dict -- requested log record
        """
        count = self.records_count(appname)
        part, index = self._map_identifier(identifier)

        if identifier > count:
            return

        if part > count - self.max_size:
            record = self._db.lindex(self._rkey % appname, index)
            return json.loads(record)

        if self.dump:
            path = j.sals.fs.join_paths(self.dump_dir, appname, "%s.msgpack" % part)
            if j.sals.fs.exists(path):
                records = msgpack.loads(j.sals.fs.read_bytes(path))
                if records and len(records) > index:
                    return json.loads(records[index])

    def remove_all_records(self, appname: str):
        """Delete all app's log records

        Arguments:
            appname {str} -- app name
        """
        self._db.delete(self._rkey % appname, self._rkey_incr % appname)
        path = j.sals.fs.join_paths(self.dump_dir, appname)

        if self.dump:
            j.sals.fs.rmtree(path)

    def tail(self, appname: str = "init", limit: int = None) -> iter:
        """Tail records

        Keyword Arguments:
            appname (str) -- appname (default: {""})
            limit (int) -- max number of record to be returned (default: {None})

        Yields:
            iter -- iterator of the requested logs
        """
        if limit:
            limit = limit - 1

        records = self._db.lrange(self._rkey % appname, 0, limit or -1)
        for record in records:
            yield json.loads(record)



import pickle


def decompress(obj):
    """dump pickle bytes object into string 
    
    Arguments:
        obj (pickle bytes) : the pickle bytes which will be dumped     
    
    Returns:
        string : the string
    """
    return pickle.loads(obj)


def compress(obj):
    """loads the data from pickle string into pickle bytes
    
    Arguments:
        obj (string) : the string
    
    Returns:
        pickle bytes : the loaded data from pickle stram
    """
    return pickle.dumps(obj)



from jumpscale.loader import j

from .base import ModelBase, JSObjBase


class ModelModel(ModelBase):
    _schema = """
    @url = model
    name* = "" (S)
    model_class = "" (S)
    """
    _name = "model"



"""
Hierarchal configurations and factories for any `jumpscale.core.base.meta.Base` type/class.

It is implemented using:

- `Factory`: the base for all factories, implements an in-memory factory for any type.
- `StoredFactory`: a factory which stores and load configuration according to the current configured store backend.


Default store configured is file system, you can show current store config by doing:

```
✗ jsctl config get store
store = filesystem
```

Get all stores configurations:

```
✗ jsctl config get stores
redis.hostname = "localhost"
redis.port = 6379
filesystem.path = "/home/abom/.config/jumpscale/secureconfig"
whoosh.path = "/home/abom/.config/jumpscale/whoosh_indexes"
```

For example, set current store to redis:

```
jsctl config update store redis
```

This will use current redis config (hostname: localhost, port: 6379).
"""
from functools import partial
from jumpscale.core import config, events

from .events import InstanceCreateEvent, InstanceDeleteEvent
from .store import ConfigNotFound, KEY_FIELD_NAME, Location
from .store.filesystem import FileSystemStore
from .store.redis import RedisStore
from .store.whooshfts import WhooshStore


STORES = {"filesystem": FileSystemStore, "redis": RedisStore, "whoosh": WhooshStore}


class DuplicateError(Exception):
    """
    raised when you try to create an instance by the same name of an existing one for a factory
    """


class Factory:
    """
    Base factory, where you can create/get/list/delete new instances.

    All of the operations are done in memory, also instances will be available as properties.

    Example:

    ```python

    class Car(Base):
        name = fields.String()
        color = fields.String()

    cars = Factory(Car)
    acar = cars.get("mycar")
    acar.name = "fiat"

    print(cars.mycar.name)
    ```
    """

    def __init__(self, type_, name_=None, parent_instance_=None, parent_factory_=None):
        """
        get a new factory given the type to create instances for.

        Any factory can have a name, parent `Base` instance and a parent factory.

        Args:
            type_ (Base): `Base` class type
            name_ (str, optional): factory name. Defaults to None.
            parent_instance_ (Base, optional): a parent `Base` instance. Defaults to None.
            parent_factory_ (Factory, optional): a parent `Factory`. Defaults to None.
        """
        self.__name = name_
        self.__parent_instance = parent_instance_
        self.__parent_factory = parent_factory_

        self.type = type_
        self.count = 0

    @property
    def parent_instance(self):
        return self.__parent_instance

    @property
    def name(self):
        return self.__name

    @property
    def parent_factory(self):
        return self.__parent_factory

    def _set_parent_factory(self, factory):
        """
        set current parent factory

        Args:
            factory (Factory)
        """
        self.__parent_factory = factory

    def find(self, name):
        """
        find an instance with the given name

        Args:
            name (str): instance name

        Raises:
            ValueError: in case the name is an internal attribute of this factory, like `get` or `new`.

        Returns:
            Base or NoneType: an instance or none
        """
        instance = getattr(self, name, None)
        if instance and not isinstance(instance, self.type):
            raise ValueError("{} is an internal attribute".format(name))
        return instance

    def _create_instance(self, name_, *args, **kwargs):
        """
        create a new instance, this method is only responsible for:

        - validating the name used is correct (must be a valid identifier and does not start with "__")
        - creating the instance from `self.type`
        - setting this instance name with the given name
        - setting the parent instance to it if this factory have a parent instance
        - update the counter and trigger `_created`

        Args:
            name_ (str): instance name

        Raises:
            ValueError: in case the name is not a valid identifier (e.g contains spaces) or starts with "__"

        Returns:
            Base: instance
        """
        if not name_.isidentifier():
            raise ValueError("{} is not a valid identifier".format(name_))

        if name_.startswith("__"):
            raise ValueError("name cannot start with '__'")

        kwargs["instance_name_"] = name_
        kwargs["parent_"] = self.parent_instance
        instance = self.type(*args, **kwargs)

        self.count += 1
        self._created(instance)
        return instance

    def new(self, name, *args, **kwargs):
        """
        get a new instance and make it available as an attribute

        Args:
            name (str): name
            *args: arbitrary arguments passed to the factory `Base` type
            *kwargs: arbitrary keyword arguments passed to the factory `Base` type

        Raises:
            DuplicateError: in case an instance with the same name exists

        Returns:
            Base: instance
        """
        if self.find(name):
            raise DuplicateError(f"instance with name {name} already exists")

        instance = self._create_instance(name, *args, **kwargs)
        setattr(self, name, instance)
        return instance

    def get(self, name, *args, **kwargs):
        """
        get an instance (will create if it does not exist)

        if the instance with `name` exists, `args` and `kwargs` are ignored.

        Args:
            name (str): name
            *args: arbitrary arguments passed to the factory `Base` type
            *kwargs: arbitrary keyword arguments passed to the factory `Base` type

        Raises:
            DuplicateError: in case an instance with the same name exists

        Returns:
            Base: instance
        """
        instance = self.find(name)
        if instance:
            return instance
        return self.new(name, *args, **kwargs)

    def _delete_instance(self, name):
        """
        delete the instance with given name

        here, we only remove the attribute

        Args:
            name (str): instance/attribute name
        """
        if hasattr(self, name):
            delattr(self, name)

    def delete(self, name):
        """
        delete an instance (with its attribute)

        this will update the count and trigger `_deleted`

        Args:
            name (str)
        """
        self.count -= 1
        self._delete_instance(name)
        self._deleted(name)

    def _deleted(self, name):
        """
        called when an instance is deleted, will trigger `InstanceDeleteEvent`

        Args:
            name (name)
        """
        event = InstanceDeleteEvent(name, factory=self)
        events.notify(event)

    def _created(self, instance):
        """
        called when an instance is created, will trigger `InstanceCreateEvent`

        Args:
            instance (Base)
        """
        event = InstanceCreateEvent(instance=instance, factory=self)
        events.notify(event)

    def list_all(self):
        """
        get a set of all instance names

        this only get a list of all properties that are of factory `Base` type.

        Returns:
            set
        """
        names = set()
        for name, value in self.__dict__.items():
            if isinstance(value, self.type) and not name.startswith("__"):
                names.add(name)
        return names


class StoredFactory(events.Handler, Factory):
    """
    Stored factories are a custom type of `Factory`, which uses current configured store backend
    to store all instance configurations.
    """

    STORE = STORES[config.get("store")]

    def __init__(self, type_, name_=None, parent_instance_=None, parent_factory_=None):
        """
        get a new stored factory given the type to create and store instances for.

        Any factory can have a name, parent `Base` instance and a parent factory.

        Once a stored factory is created, it tries to lazy-load all current configuration for given `type_`.

        Args:
            type_ (Base): `Base` class type
            name_ (str, optional): factory name. Defaults to None.
            parent_instance_ (Base, optional): a parent `Base` instance. Defaults to None.
            parent_factory_ (Factory, optional): a parent `Factory`. Defaults to None.
        """
        super().__init__(type_, name_=name_, parent_instance_=parent_instance_, parent_factory_=parent_factory_)
        self.__store = None

        if not parent_instance_:
            # no parent, then load all instance configurations
            # if it's a parent, it should trigger this loading
            self._load()

        # to handle when a parent instance is deleted
        events.add_listenter(self, InstanceDeleteEvent)

        # if we need to always reload the config from store when getting an instance
        self.always_reload = config.get("factory").get("always_reload", False)

    @property
    def parent_location(self):
        if not self.parent_factory:
            raise ValueError("cannot get parent location if parent factory is not set")
        return self.parent_factory.location

    @property
    def location(self):
        """
        get a unique location for this factory

        Returns:
            Location: location object
        """
        name_list = []

        # first, get the location of parent factory if any
        if self.parent_factory:
            name_list += self.parent_location.name_list

        # if we have a parent instance, then this location should be unique
        # for this instance
        if self.parent_instance:
            name_list.append(self.parent_instance.instance_name)

        # if this factory have a name, append it too
        if self.name:
            name_list.append(self.name)

        # then we append the location of the type
        type_location = Location.from_type(self.type)
        name_list += type_location.name_list

        return Location(*name_list, type_=self.type)

    @property
    def store(self):
        if not self.__store:
            self.__store = self.STORE(self.location)
        return self.__store

    def _validate_and_save_instance(self, instance):
        """
        validate and save a given instance to the store

        Args:
            instance (Base)
        """
        instance.validate()
        self.store.save(instance.instance_name, instance._get_data())
        if instance.parent and hasattr(instance.parent, "save"):
            instance.parent.save()

    def handle(self, ev):
        """
        handle when the parent instance is deleted

        Args:
            ev (InstanceDeleteEvent): instance delete event
        """
        # do nothing if this is a root factory
        if not self.parent_instance or not self.parent_factory:
            return

        # handle deletion of children
        if isinstance(ev, InstanceDeleteEvent):
            if ev.factory == self.parent_factory and ev.name == self.parent_instance.instance_name:
                for name in self.list_all():
                    self.delete(name)

    def _load_sub_factories(self, instance):
        """
        load sub-factories for an instance

        this will also set current factory as a parent for any sub-factories this instance have

        Args:
            instance (Base)
        """
        for factory in instance._get_factories().values():
            factory._set_parent_factory(self)
            factory._load()

    def _init_save_and_sub_factories(self, instance):
        """
        allow an instance to have a `save` method

        also, this method inits the loading of sub-factories for this instance

        Args:
            instance (Base)
        """
        instance.save = partial(self._validate_and_save_instance, instance)
        self._load_sub_factories(instance)

    def _get_object_from_config(self, name, data):
        instance = self._create_instance(name, **data)
        self._init_save_and_sub_factories(instance)
        return instance

    def _load_from_store(self, name):
        """
        loads instance from store

        Args:
            name (str): instance name
        Returns:
            Base or NoneType: an instance or none
        """
        try:
            instance_config = self.store.get(name)
        except ConfigNotFound:
            return

        instance = self._get_object_from_config(name, instance_config)
        setattr(self, name, instance)
        return instance

    def find(self, name):
        """
        find an instance with the given name

        Args:
            name (str): instance name

        Raises:
            ValueError: in case the name is an internal attribute of this factory, like `get` or `new`.

        Returns:
            Base or NoneType: an instance or none
        """
        instance = super().find(name)
        if instance:
            if self.always_reload:
                try:
                    instance._set_data(self.store.get(name))
                except ConfigNotFound:
                    # no config is written, still not saved
                    pass
        else:
            instance = self._load_from_store(name)

        return instance

    def new(self, name, *args, **kwargs):
        """
        get a new instance and make it available as an attribute

        this method also initialize sub-factories for this instance.

        Args:
            name (str): name
            *args: arbitrary arguments passed to the factory `Base` type
            *kwargs: arbitrary keyword arguments passed to the factory `Base` type

        Raises:
            DuplicateError: in case an instance with the same name exists

        Returns:
            Base: instance
        """
        instance = super().new(name, *args, **kwargs)
        self._init_save_and_sub_factories(instance)
        return instance

    def get_instance_property(self, name):
        """
        return a new property descriptor for a given name, this will help in lazy-loading
        as this property will only create and load an instance from the store once accessed.

        Args:
            name (str): instance name

        Returns:
            property: property descriptor (object)
        """
        inner_name = f"__{name}"

        def getter(factory):
            # we need to avoid AttributeError which hasattr swallows
            # instead we check the internal __dict__
            if inner_name in factory.__dict__:
                return getattr(factory, inner_name)

            instance = self._get_object_from_config(name, factory.store.get(name))
            setattr(factory, inner_name, instance)
            return instance

        return property(getter)

    def _load(self):
        """
        lazy-load all instance configuration

        it will a new class for this factory with property descriptors for all instances listed by the store
        """
        # get a class name
        cls_name = self.location.name.replace(".", "")
        # get a new class based on self.__class__
        new_cls = type(cls_name, (self.__class__,), {})

        # list all instance names in the store and set a property descriptor as an attribute
        for name in self.store.list_all():
            # it will create the actual instance with configuration from the store when accessed only
            setattr(new_cls, name, self.get_instance_property(name))

        # update current factory class with this class
        self.__class__ = new_cls

    def _delete_instance(self, name):
        """
        delete an instance

        this will delete it from the store too, also, delete its property descriptor if found

        Args:
            name (str): instance name
        """
        self.store.delete(name)

        class_prop = getattr(self.__class__, name, None)
        if isinstance(class_prop, property):
            # created with a property descriptor inside the class, delete it
            delattr(self.__class__, name)

            # check if the instance was actually created or not (if the property was accessed)
            inner_name = f"__{name}"
            if hasattr(self, inner_name):
                # delete the instance itself too if it was accessed and created
                delattr(self, inner_name)
        else:
            super()._delete_instance(name)

    def find_many(self, cursor_=None, limit_=None, **query):
        """
        do a search against the store (not loaded objects) with this query.

        queries can relate to current store backend used.

        Keyword Args:
            cursor_ (any, optional): an optional cursor, to start searching from. Defaults to None.
            limit_ (int, optional): results limit. Defaults to None.
            query: a mapping for field/query, e.g. first_name="aa"

        Returns:
            tuple: the new cursor, total results count, and a list of objects as a result
        """
        if not query:
            raise ValueError("at least one query parameter is required, e.g. age=10")

        new_cursor, count, result = self.store.find(cursor_=cursor_, limit_=limit_, **query)
        return new_cursor, count, (self._get_object_from_config(data[KEY_FIELD_NAME], data) for data in result)

    def list_all(self):
        """
        get all instance names (stored or not)

        Returns:
            set: instance names
        """
        names = set(self.store.list_all())
        return names.union(super().list_all())

    def __iter__(self):
        for value in vars(self).values():
            if isinstance(value, self.type):
                yield value

    def __eq__(self, other):
        return self.location == other.location

    def __hash__(self):
        return hash(self.location)



from whoosh import fields
from whoosh.index import create_in, exists_in, open_dir
from whoosh.qparser import FuzzyTermPlugin, GtLtPlugin, MultifieldParser, PhrasePlugin
from whoosh.writing import AsyncWriter

from . import ConfigNotFound, EncryptedConfigStore, KEY_FIELD_NAME
from .serializers import Serializer

from jumpscale.sals.fs import join_paths, mkdirs

# a map betwen our indexable fields and whoosh fields
# for now we don't support nested fields like Lists or Objects
# they will only be stored but not indexed
FIELD_MAP = {
    "Boolean": fields.BOOLEAN(stored=True),
    "Bytes": fields.TEXT(stored=True),
    "Email": fields.TEXT(stored=True),
    "GUID": fields.TEXT(stored=True),
    "IPAddress": fields.TEXT(stored=True),
    "IPRange": fields.TEXT(stored=True),
    "Json": fields.TEXT(stored=True),
    "Path": fields.ID(stored=True),
    "String": fields.TEXT(stored=True),
    "Tel": fields.TEXT(stored=True),
    "URL": fields.TEXT(stored=True),
    "Integer": fields.NUMERIC(bits=64, stored=True, sortable=True),
    "Float": fields.NUMERIC(float, bits=64, stored=True, sortable=True),
    "Port": fields.NUMERIC(stored=True, sortable=True),
    "Date": fields.NUMERIC(stored=True, sortable=True),
    "DateTime": fields.NUMERIC(stored=True, sortable=True),
    "Time": fields.NUMERIC(stored=True, sortable=True),
}

# enums and secret fields are special case,
# for enum, it can be stored as int, bool, or str values...
# so, the field type is selected based on enum values (assuming they're all of the same type)
ENUM_FIELD = "Enum"
# for secret, its field name prefixed with underscores is not valid in whoosh
# they are handled when reading/writing the data
SECRET_FIELD = "Secret"


class WhooshStore(EncryptedConfigStore):
    """
    whoosh store is an EncryptedConfigStore

    It saves and indexes the data in a whoosh index
    """

    def __init__(self, location):
        """
        create a new redis store, the location given will be used to generate keys

        this keys will be combined to get/set instance config

        Args:
            location (Location)
        """
        super().__init__(location, Serializer())
        config = self.config_env.get_store_config("whoosh")
        self.base_index_path = config["path"]

        self.schema = self.get_schema()
        self.index = self.get_index(self.schema)

        self.default_plugins = [FuzzyTermPlugin(), GtLtPlugin(), PhrasePlugin()]
        self.default_pagenum = 1
        self.default_pagelen = 20

    @property
    def index_path(self):
        path = join_paths(self.base_index_path, self.location.name)
        mkdirs(path)
        return path

    @property
    def type_fields(self):
        return self.location.type._fields.items()

    def get_schema(self):
        schema_fields = {
            KEY_FIELD_NAME: fields.ID(unique=True, stored=True),
        }

        for name, field in self.type_fields:
            field_type_name = field.__class__.__name__
            if field_type_name in FIELD_MAP:
                schema_fields[name] = FIELD_MAP[field_type_name]
            elif field_type_name == ENUM_FIELD:
                # it depends on the type of enum values
                enum_value_type = type(field.default)
                if isinstance(enum_value_type, (str, bytes, bytearray)):
                    schema_field = fields.TEXT(stored=True)
                elif isinstance(enum_value_type, (int, float)):
                    schema_field = fields.NUMERIC(stored=True)
                else:
                    schema_field = fields.STORED

                schema_fields[name] = schema_field
            else:
                schema_fields[name] = fields.STORED

        return fields.Schema(**schema_fields)

    def get_index(self, schema):
        if exists_in(self.index_path):
            return open_dir(self.index_path, schema=schema)
        return create_in(self.index_path, schema=schema)

    def get_reader(self):
        return self.index.reader()

    def get_writer(self):
        return AsyncWriter(self.index)

    def get_searcher(self, up_to_date=True):
        searcher = self.index.searcher()

        if up_to_date and not searcher.up_to_date():
            searcher = searcher.refresh()

        return searcher

    def read(self, instance_name):
        with self.get_searcher() as searcher:
            kw = {KEY_FIELD_NAME: instance_name}
            doc = searcher.document(**kw)

            if not doc:
                raise ConfigNotFound(f"cannot find config for {instance_name} in the index")

            for name, field in self.type_fields:
                # whoosh does not store None values, so, we just set them
                # if they are not set, that means when they're added, they'd the value of None
                if name not in doc and field.stored:
                    doc[name] = None

                # add __ to field name by hand
                # as we cannot add a field that starts with "__" in whoosh schema
                if field.__class__.__name__ == SECRET_FIELD:
                    name_with_prefix = f"__{name}"
                    doc[name_with_prefix] = doc[name]
                    doc.pop(name)

            return doc

    def write(self, instance_name, data):
        data[KEY_FIELD_NAME] = instance_name

        for name, field in self.type_fields:
            if field.__class__.__name__ == SECRET_FIELD:
                name_with_prefix = f"__{name}"
                if name_with_prefix in data:
                    # remove "__", as the whoosh field cannot start with it
                    # the original field name is the one used in whoosh schema
                    data[name] = data[name_with_prefix]
                    data.pop(name_with_prefix)

        writer = self.get_writer()
        writer.update_document(**data)
        writer.commit()

    def list_all(self):
        with self.get_reader() as reader:
            for _, doc in reader.iter_docs():
                yield doc[KEY_FIELD_NAME]

    def find(self, cursor_=None, limit_=None, **queries):
        fields = queries.keys()
        query_text = " ".join([f"{field}:{queries[field]}" for field in fields])

        parser = MultifieldParser(fields, schema=self.schema)
        parser.add_plugins(self.default_plugins)

        query = parser.parse(query_text)
        searcher = self.get_searcher()

        if not cursor_:
            cursor_ = self.default_pagenum
        if not limit_:
            limit_ = self.default_pagelen

        result = searcher.search_page(query, pagenum=cursor_, pagelen=limit_)

        if result.is_last_page():
            new_cursor = None
        else:
            new_cursor = result.pagenum + 1

        if result.total >= limit_:
            result = result[:limit_]

        return new_cursor, len(result), (hit for hit in result)

    def delete(self, instance_name):
        writer = self.get_writer()
        writer.delete_by_term(KEY_FIELD_NAME, instance_name)
        writer.commit()



"""Console module helps with coloring in the console and asking for input from the user.
```
JS-NG> j.tools.console.printcolors("{RED}Hello{BGRED}What{WHITE}OK")                                                                     
JS-NG> j.tools.console.printcolors("{RED}Hello{BGRED}What{RESET}{WHITE}OK")                                                              
```

"""


import colorama
import getpass
import os


NAMES_TO_COLORS = {}
for attrname in dir(colorama.Fore):
    if attrname.isupper():
        NAMES_TO_COLORS[attrname] = getattr(colorama.Fore, attrname)

for attrname in dir(colorama.Back):
    if attrname.isupper():
        NAMES_TO_COLORS["BG" + attrname] = getattr(colorama.Back, attrname)

NAMES_TO_COLORS["RESET"] = colorama.Style.RESET_ALL


def format(s):
    return s.format(**NAMES_TO_COLORS)


def printcolors(s):
    """
	>>> j.tools.console.printcolors("{RED}Hello world")
	Hello world
	>>> j.tools.console.printcolors("{GREEN}Hello world")
	Hello world

	Arguments:
	s {[type]} -- [description]
	"""
    print(format(s))


def ask_password(prompt="Password : ", forbiddens=[]):
    """Prompt the user for a password without echoing
    
    Keyword Arguments:
        prompt {str} -- the question message (default: {"Password : "})
        forbiddens {list} -- the list of bad passwords (default: {[]})
    
    Returns:
        str -- the appropriate input password
    """
    password = getpass.getpass(prompt)
    if password not in forbiddens:
        return password
    else:
        return ask_password(prompt, forbiddens)


def ask_yes_no(prompt="[y/n] :", default="y", valid=["y", "n"]):
    """Display a yes/no question and loop until a valid answer is entered
    
    Keyword Arguments:
        prompt {str} -- the question message (default: {'[y/n] :'})
        default {str} -- the default answer if there is no answer (default: {"y"})
        valid {list} -- the list of appropriate answers (default: {["y", "n"]})
    
    Returns:
        str -- the answer
    """

    answer = input(prompt)
    if answer in valid:
        return answer
    elif answer == "":
        return default
    else:
        return ask_yes_no(prompt, default, valid)


def ask_int(prompt="Type int :"):
    try:
        return int(input(prompt))
    except ValueError:
        return ask_int(prompt)


def ask_int_in_range(mini, maxi, prompt="Type int :"):
    """Get an integer response between two integer on asked question
    
    Arguments:
        mini {int} -- the minimum value for the number
        maxi {int} -- the maximum value for the number
    
    Keyword Arguments:
        prompt {str} -- the question message (default: {"Type int :"})
    
    Returns:
        int -- the input number on the range provided
    """
    try:
        answer = int(input(prompt))
        if mini <= answer <= maxi:
            return answer
        else:
            return ask_int_in_range(mini, maxi, prompt)
    except ValueError:
        return ask_int_in_range(mini, maxi, prompt)


def ask_float(prompt="Type float :"):
    try:
        return float(input(prompt))
    except ValueError:
        return ask_float(prompt)


def ask_float_in_range(mini, maxi, prompt="Type float :"):
    """Get an float response between two float on asked question
    
    Arguments:
        mini {float} -- the minimum value for the number
        maxi {float} -- the maximum value for the number
    
    Keyword Arguments:
        prompt {str} -- the question message (default: {"Type float :"})
    
    Returns:
        float -- the input number on the range provided
    """
    try:
        answer = float(input(prompt))
        if mini <= answer <= maxi:
            return answer
        else:
            return ask_float_in_range(mini, maxi, prompt)
    except ValueError:
        return ask_float_in_range(mini, maxi, prompt)


def _print_choices(choices_list):
    """Helper function : clear screen and print the choices in numbers"""
    os.system("clear")
    number = 0
    for choice in choices_list:
        number += 1
        print(f"{number}. " + choice)


def ask_choice(prompt="Type choice number : ", choices_list=[]):
    """Get an option from provided list
    
    Keyword Arguments:
        prompt {str} -- the question message (default: {"Type choice number : "})
        choices_list {list} -- the available options (default: {[]})
    
    Returns:
        str -- the chosen option
    """
    _print_choices(choices_list)
    answer = input(prompt)
    try:
        return choices_list[int(answer) - 1]
    except (IndexError, ValueError):
        return ask_choice(prompt, choices_list)


def ask_multi_choices(prompt="Add to choices : ", choices_list=[], to_save="s", to_quit="q"):
    """Collect multi choices from list
    
    Keyword Arguments:
        prompt {str} -- the question method (default: {"Add to choices : "})
        choices_list {list} -- the available options (default: {[]})
        to_save {str} -- escape and save choices (default: {"s"})
        to_quit {str} -- escape without saving (default: {"q"})
    
    Returns:
        list -- list of the selected choices
    """
    selected_choices = []
    print(f"'{to_save}' to save and '{to_quit}' to quit")
    _print_choices(choices_list)

    while True:
        answer = input(prompt)
        if answer == to_quit:
            return []
        elif answer == to_save or answer == "":
            return selected_choices
        else:
            try:
                selected_choices.append(choices_list[int(answer) - 1])
            except (IndexError, ValueError):
                return ask_multi_choices(prompt, choices_list, to_save, to_quit)


def ask_multi_lines(prompt="Type :", escape_string="."):
    """Get input from user provided multilines
    
    Keyword Arguments:
        prompt {str} -- the question message (default: {"Type :"})
        escape_string {str} -- escape character (default: {"."})
    
    Returns:
        str -- the text seperated by lines
    """
    text = []
    user_input = input(prompt)
    while user_input != escape_string:
        text.append(user_input)
        user_input = input(prompt)
    return "\n".join(text)


def ask_string(prompt="Type :"):
    """Just input function
    
    Keyword Arguments:
        prompt {str} -- the question message (default: {"Type :"})
    
    Returns:
        str -- the string input
    """
    return input(prompt)

    print(format(s))


def printobj(obj):
    from pprint import pprint

    pprint(obj)



"""This module is providing everything needed for decent filesystem management
# Using System Fs

Manipulating filesystem is one of the most common things in the daily life of a developer, administrator, .. etc. js-ng comes with tons of helpers and utilities around that in `j.sals.fs`. You will find the module already self documented with lots of examples in each function.


### Get current working dir

```
> j.sals.fs.cwd()
'/home/xmonader/wspace/threefoldtech/js-ng'
```

### Get basename
```
> j.sals.fs.basename(j.sals.fs.cwd())
'js-ng'
```

### Get Dir name
```
> j.sals.fs.dirname(j.sals.fs.cwd())
'/home/xmonader/wspace/threefoldtech'
> j.sals.fs.parent(j.sals.fs.cwd())
'/home/xmonader/wspace/threefoldtech'
```

### is dir

```
> j.sals.fs.is_dir(path= '/home/rafy')
True
```
### is file
```
> j.sals.fs.is_file(path= '/home/rafy')
False
```

### is ascii file

```
> j.sals.fs.is_ascii_file(path="/home/rafy/testfile")
True
```

### Is absolute path

```
> j.sals.fs.is_absolute('/home/rafy/')
True
```

### Check if empty dir
```
> j.sals.fs.is_empty_dir("/home/rafy/empty_dir")
True
```

### File paths exists or not

```
> j.sals.fs.exists("/home/rafy/testing_make_dir/test1")
True
```

## Reading/Writing to a file

There're some helpers around reading/writing text, binary like `read_text`, `read_binary`, `read_file`, `write_text`, `write_binary`, `touch`

### Touching a new file

```
> j.sals.fs.touch("/home/rafy/testing_touch")
```

### Reading a text

```
> j.sals.fs.read_text("/home/rafy/testing_text.txt")
'hello world\n'

```

### Reading binary

```
> j.sals.fs.read_bytes("/home/rafy/testing_text.txt")
b'hello world\n'
```


### Writing text

```
> j.sals.fs.write_text(path="/home/rafy/testing_text.txt",data="hello world")
11
```

### Writing binary

```
> j.sals.fs.write_bytes(path="/home/rafy/testing_text.txt",data=b"hello world")
11
```


rename, move, copy_file, copy_tree, mkdir, mkdirs, join_paths, , rmtree, rm_empty_dir, symlink, chmod, chown, basename, dirname, normalizing paths, expanding `~`



### Making directories
```
> j.sals.fs.mkdirs("/home/rafy/testing_make_dir/test1/test2",exist_ok=False)
```
will raise if in case the directory already exists

```
> j.sals.fs.mkdirs("/home/rafy/testing_make_dir/test1/test2",exist_ok=True)
```
Won't raise if the directory exists

### Get the stem of the filepath

```
> j.sals.fs.stem("/tmp/tmp-5383p1GOmMOOwvfi.tpl") 'tmp-5383p1GOmMOOwvfi'

```

### Get the parent

```
> j.sals.fs.parent("/home/rafy/testing_make_dir/test1")
'/home/rafy/testing_make_dir'
```

### Get parents

```
> j.sals.fs.parents("/tmp/home/ahmed/myfile.py")
    [PosixPath('/tmp/home/ahmed'),
    PosixPath('/tmp/home'),
    PosixPath('/tmp'),
    PosixPath('/')]
```


### Rename file
```
> j.sals.fs.rename("/home/rafy/testing_make_dir","/home/rafy/testing_dir")
```

### Expand user
```
> j.sals.fs.expanduser("~/work")
'/home/xmonader/work'
```

### Get temporary filename
```
> j.sals.fs.get_temp_filename(dir="/home/rafy/")
'/home/rafy/tmp6x7w71ml'
```

```
> j.sals.fs.get_temp_dirname(dir="/home/rafy")
'/home/rafy/tmpntm2ptqy'
```

```
> j.sals.fs.join_paths("home","rafy")
'home/rafy'
```

### Resolving a path
```
> j.sals.fs.resolve("")
PosixPath('/home/rafy/Documents')
> j.sals.fs.resolve("./testing_text.txt")
PosixPath('/home/rafy/Documents/testing_text.txt')
```

## Walkers
It's very common to walk on the filesystem and filtering based on some properties of the path
And very fancy walkers

### Walk over on files only

```
for el in walk('/tmp', filter_fun=j.sals.fs.is_file) : ..
```
or more specific function walk_files

```
for el in walk_files('/tmp') : ..
```



### Walk over on dirs only

```
for el in walk('/tmp', filter_fun=j.sals.fs.is_dir) : ..
```
or more specific function `walk_dirs`

```
for el in walk_dirs('/tmp') : ..

```

### walk over with a bit complex filter

 Walk over paths that are files or dirs and longer than 4 characters in the name

```
for el in walk('/tmp', filter_fun= lambda x: len(x)>4 and (j.sals.fs.is_file(x) or j.sals.fs.is_dir(x)) ) : ..
```


There are more functionality available in the SAL `j.sals.fs` make sure you check the API documentation for more.

"""

import pathlib
import tempfile
import os
import shutil
import stat
from distutils import dir_util
from typing import List

basename = os.path.basename
dirname = os.path.dirname
common_path = os.path.commonpath
common_prefix = os.path.commonprefix
norm_path = os.path.normpath
norm_case = os.path.normcase
get_access_time = os.path.getatime
get_modified_time = os.path.getmtime
get_creation_time = os.path.getctime
sep = os.path.sep
is_samefile = os.path.samefile
expandvars = os.path.expandvars
expanduser = os.path.expanduser
realpath = os.path.realpath


def home() -> str:
    """Return the home directory
    e.g
        j.sals.fs.home()  -> '/home/rafy'

    Returns:
        str: home directory.
    """
    return str(pathlib.Path.home())


def cwd() -> str:
    """Return current working directory.
    e.g
        j.sals.fs.cwd()  -> '/home/rafy'


    Returns:
        str: current directory
    """
    return str(pathlib.Path.cwd())


def is_dir(path: str) -> bool:
    """Checks if path is a dir
    e.g
        j.sals.fs.is_dir(path= '/home/rafy')  -> True
        j.sals.fs.is_dir(path= '/home/rafy/file.txt')  -> False

    Args:
        path (str): path to check if is directory
    Returns:
        bool: True if is dir and False otherwise
    """
    return pathlib.Path(path).is_dir()


def is_file(path: str) -> bool:
    """Checks if path is a file
    e.g
        j.sals.fs.is_file(path= '/home/rafy')  -> False
        j.sals.fs.is_file(path= '/home/rafy/file.txt')  -> True

    Args:
        path (str): path to check if is file

    Returns:
        bool: True if is file and False otherwise
    """
    return pathlib.Path(path).is_file()


def is_symlink(path: str) -> bool:
    """Checks if path symlink
    e.g
        j.sals.fs.is_symlink('/home/rafy/testfile3')  -> True
        j.sals.fs.is_symlink('/home/rafy/testfile2')  -> False

    Args:
        path (str): path to check if symlink

    Returns:
        bool: True if symlink False otherwise
    """
    return pathlib.Path(path).is_symlink()


def is_absolute(path: str) -> bool:
    """Checks if path is absolute
    e.g
        j.sals.fs.is_absolute('/home/rafy/')  -> True
        j.sals.fs.is_absolute('~/rafy/')  -> False

    Args:
        path (str): path to check if it is absolute

    Returns:
        bool: True if absolute
    """
    return pathlib.Path(path).is_absolute()


def is_mount(path: str) -> bool:
    # TODO add example here
    """Checks if path is mount

    Args:
        path (str): path to check if it is mounded or not

    Returns:
        bool: True if mount
    """

    return pathlib.Path(path).is_mount()


def is_ascii_file(path: str, checksize=4096) -> bool:
    """Checks if file `path` is ascii
    e.g
         j.sals.fs.is_ascii_file(path="/home/rafy/testfile")  -> True

    Args:
        path (str): file path
        checksize (int, optional): checksize. Defaults to 4096.

    Returns:
        bool: True if ascii file
    """
    # TODO: let's talk about checksize feature.
    try:
        with open(path, encoding="ascii") as f:
            f.read()
            return True
    except UnicodeDecodeError:
        return False


def is_empty_dir(path: str) -> bool:
    """Checks if path is emptry directory
    e.g
        j.sals.fs.is_empty_dir("/home/rafy/empty_dir")  -> True
        j.sals.fs.is_empty_dir("/home/rafy")  -> False

    Args:
        path (str): path to check if empty directory

    Returns:
        bool: True if path is emptry directory
    """

    try:
        g = pathlib.Path(path).iterdir()
        next(g)
    except StopIteration:
        # means we can't get next entry -> dir is empty.
        return True
    else:
        return False


is_binary_file = lambda path: not is_ascii_file(path)


def is_broken_link(path: str, clean=False) -> bool:
    """Checks if path is a broken symlink

    Args:
        path (str): path to check
        clean (bool, optional): remove symlink if broken. Defaults to False.

    Raises:
        NotImplementedError: [description]

    Returns:
        bool: True if path is a broken symlink
    """
    raise NotImplementedError()


def stem(path: str) -> str:
    """returns the stem of a path (path without parent directory and without extension)
    e.g
        j.sals.fs.stem("/tmp/tmp-5383p1GOmMOOwvfi.tpl")  -> 'tmp-5383p1GOmMOOwvfi'

    Args:
        path (str): path we want to get its stem

    Returns:
        str: path without parent directory and without extension
    """
    return pathlib.Path(path).stem


def mkdir(path: str, exist_ok=True):
    """Makes directory at path
    e.g
        j.sals.fs.mkdir("/home/rafy/testing_make_dir")
        j.sals.fs.mkdir("/home/rafy/testing_make_dir",exist_ok=True)
        j.sals.fs.mkdir("/home/rafy/testing_make_dir",exist_ok=False) -> File exists: '/home/rafy/testing_make_dir' (raise exception as the file exist and  exist_ok flag is false)

    Args:
        path (str): path to create dir at
        exist_ok (bool, optional): won't fail if directory exists. Defaults to True.

    Returns:
        [type]: [description]
    """
    return pathlib.Path(path).mkdir(exist_ok=exist_ok)


def mkdirs(path: str, exist_ok=True):
    """Creates dir as well as all non exisitng parents in the path
    e.g
        j.sals.fs.mkdirs("/home/rafy/testing_make_dir/test1/test2",exist_ok=False)
        j.sals.fs.mkdirs("/home/rafy/testing_make_dir/test1/test2",exist_ok=True)
        j.sals.fs.mkdirs("/home/rafy/testing_make_dir/test1/test2",exist_ok=False)  -> File exists: '/home/rafy/testing_make_dir/test1/test2'(raise exception as the file exist and  exist_ok flag is false)

    Args:
        path (str): path to create dir at
        exist_ok (bool, optional): won't fail if directory exists. Defaults to True.
    """
    return os.makedirs(path, exist_ok=exist_ok)


def make_path(path):
    """
    to ensure the given path, create it if it does not exist

    Args:
        path (str): path
    """
    if not os.path.exists(path):
        os.makedirs(os.path.dirname(path), exist_ok=True)
        return pathlib.Path(path).touch()



def parent(path: str) -> str:
    """Get path's parent
    e.g
        j.sals.fs.parent("/home/rafy/testing_make_dir/test1")  -> '/home/rafy/testing_make_dir'

    Args:
        path (str): path to get its parent

    Returns:
        str: parent path.
    """
    return str(pathlib.Path(path).parent)


def parents(path: str) -> List[str]:
    """Get parents list

    e.g
        j.sals.fs.parents("/tmp/home/ahmed/myfile.py") ->
    [PosixPath('/tmp/home/ahmed'),
    PosixPath('/tmp/home'),
    PosixPath('/tmp'),
    PosixPath('/')]

    Args:
        path (str): path to get its parents

    Returns:
        List[str]: list of parents
    """

    return list([str(p) for p in pathlib.Path(path).parents])


def path_parts(path: str) -> List[str]:
    """Convert path to a list of parts
    e.g
        '/tmp/tmp-5383p1GOmMOOwvfi.tpl' ->  ('/', 'tmp', 'tmp-5383p1GOmMOOwvfi.tpl')
    Args:
        path (str): path to convert to parts

    Returns:
        List[str]: path parts.
    """
    return pathlib.Path(path).parts


def exists(path: str) -> bool:
    """Checks if path exists
    e.g
        j.sals.fs.exists("/home/rafy/testing_make_dir/test1")  -> True
        j.sals.fs.exists("/home/rafy/testing_make_dir/fasdljd")  -> False

    Args:
        path (str): path to check for existence

    Returns:
        bool: True if exists
    """
    return pathlib.Path(path).exists()


def rename(path1: str, path2: str):
    """Rename path1 to path2
    e.g
        j.sals.fs.rename("/home/rafy/testing_make_dir","/home/rafy/testing_dir")

    Args:
        path1 (str): source path
        path2 (str): dest path

    """
    return pathlib.Path(path1).rename(path2)


def expanduser(path: str) -> str:
    """Expands the tilde `~` to username
    e.g
        j.sals.fs.expanduser("~/work") -> '/home/xmonader/work'
    Args:
        path (str): path with optionally `~`

    Returns:
        str: path with tilde `~` resolved.
    """
    return str(pathlib.Path(path).expanduser())


def unlink(path: str):
    """unlink path
    e.g
        j.sals.fs.unlink("/home/rafy/testfile3")
    Args:
        path (str): path to unlink
    """
    return pathlib.Path(path).unlink()


def read_text(path: str) -> str:
    """read ascii content at `path`
    e.g
        j.sals.fs.read_text("/home/rafy/testing_text.txt")  -> 'hello world\n'

    Args:
        path (str): ascii file path

    Returns:
        str: ascii content in path
    """
    return pathlib.Path(path).read_text()


read_ascii = read_file = read_text


def read_bytes(path: str) -> bytes:
    """read binary content at `path`
    e.g
        j.sals.fs.read_bytes("/home/rafy/testing_text.txt")  -> b'hello world\n'

    Args:
        path (str): binary file path

    Returns:
        bytes: binary content in path
    """
    return pathlib.Path(path).read_bytes()


read_binary = read_file_binary = read_bytes


def write_text(path: str, data: str, encoding=None):
    """write text `data` to path `path` with encoding
    e.g
        j.sals.fs.write_text(path="/home/rafy/testing_text.txt",data="hello world")  -> 11

    Args:
        path (str): path to write to
        data (str): ascii content
        encoding ([type], optional): encoding. Defaults to None.

    Returns:
        int: Return the decoded contents of the pointed-to file as a string

    """
    return pathlib.Path(path).write_text(data, encoding)


write_ascii = write_file = write_text


def write_bytes(path: str, data: bytes):
    """write binary `data` to path `path`
    e.g
        j.sals.fs.write_bytes(path="/home/rafy/testing_text.txt",data=b"hello world")  -> 11

    Args:
        path (str): path to write to
        data (bytes): binary content

    Returns:
        int: Return the binary contents of the pointed-to file as a bytes object
    """
    return pathlib.Path(path).write_bytes(data)


write_binary = write_file_binary = write_bytes


def touch(path: str):
    """create file
    e.g
        j.sals.fs.touch("/home/rafy/testing_touch")

    Args:
        path (str): path to create file

    """
    return pathlib.Path(path).touch()


def get_temp_filename(mode="w+b", buffering=-1, encoding=None, newline=None, suffix=None, prefix=None, dir=None) -> str:
    """Get temp filename
    e.g
        j.sals.fs.get_temp_filename(dir="/home/rafy/")  -> '/home/rafy/tmp6x7w71ml'

    Args:
        mode (str, optional): [description]. Defaults to "w+b".
        buffering (int, optional): buffering. Defaults to -1.
        encoding ([type], optional): encoding . Defaults to None.
        newline ([type], optional):  Defaults to None.
        suffix ([type], optional): ending suffix. Defaults to None.
        prefix ([type], optional): prefix . Defaults to None.
        dir ([type], optional): where to create the file. Defaults to None.

    Returns:
        [str]: temp filename
    """
    return tempfile.NamedTemporaryFile(mode, buffering, encoding, newline, suffix, prefix, dir).name


def get_temp_dirname(suffix=None, prefix=None, dir=None) -> str:
    """Get temp directory name
    e.g
        j.sals.fs.get_temp_dirname(dir="/home/rafy")  -> '/home/rafy/tmpntm2ptqy'

    Args:
        suffix ([type], optional): ending suffix. Defaults to None.
        prefix ([type], optional): prefix . Defaults to None.
        dir ([type], optional): where to create the directory. Defaults to None.


    Returns:
        str: temp directory name.
    """
    return tempfile.TemporaryDirectory(suffix, prefix, dir).name


NamedTemporaryFile = tempfile.NamedTemporaryFile
TempraryDirectory = tempfile.TemporaryDirectory
mkdtemp = tempfile.mkdtemp
mkstemp = tempfile.mkstemp
get_temp_dir = tempfile.gettempdir


def parts_to_path(parts: List[str]) -> str:
    """Convert list of path parts into a path string
    e.g
        j.sals.fs.parts_to_path(["home","rafy"])  -> 'home/rafy'

    Args:
        parts (List[str]): path parts

    Returns:
        str: joined path parts
    """
    path = pathlib.Path(parts[0])
    for p in parts[1:]:
        path = path.joinpath(p)
    return str(path)


def join_paths(*paths) -> str:
    """
    Convert tuple of path parts into a path string
    e.g
        j.sals.fs.join_paths("home","rafy")  -> 'home/rafy'

    Args:
        parts (tuple): path parts (path parts comma seprated and they will be used as a tuple)

    Returns:
        str: joined path parts

    """

    return parts_to_path(paths)


def rm_emptry_dir(path: str):
    """Remove empty directory if the directory is not empty it will throw exception (Directory not empty)
    e.g
        j.sals.fs.rm_emptry_dir("/home/rafy/empty_dir")

    Args:
        path (str): path to remove.
    """
    path = pathlib.Path(path)
    path.rmdir()


def rmtree(path: str):
    """Remove directory tree
    Args:
        path (str): path to remove
    """
    path = pathlib.Path(path)
    if path.is_file() or path.is_symlink():
        os.remove(path)
    elif path.is_dir():
        shutil.rmtree(path)


def copy_stat(src: str, dst: str, times=True, perms=True):
    """Copy stat of src to dst

    Args:
        src (str): source path
        dst (str): destination
        times (bool, optional):  Defaults to True.
        perms (bool, optional): permissions Defaults to True.
    """
    st = os.stat(src)
    if hasattr(os, "utime"):
        os.utime(dst, (st.st_atime, st.st_mtime))
    if hasattr(os, "chmod"):
        m = stat.S_IMODE(st.st_mode)
        os.chmod(dst, m)


def copy_file(src: str, dst: str, times=False, perms=False):
    """Copy the file, optionally copying the permission bits (mode) and
        last access/modify time. If the destination file exists, it will be
        replaced. Raises OSError if the destination is a directory. If the
        platform does not have the ability to set the permission or times,
        ignore it.
        This is shutil.copyfile plus bits of shutil.copymode and
        shutil.copystat's implementation.
        shutil.copy and shutil.copy2 are not supported but are easy to do.
    e.g
        j.sals.fs.copy_file(src="/home/rafy/testing_text.txt",dst="/home/rafy/Documents/testing_text.txt")

    Args:
        src (str): source path
        dst (str): destination

    """
    shutil.copyfile(src, dst)
    if times or perms:
        copy_stat(src, dst, times, perms)


def symlink(src: str, dst: str, overwrite=False):
    """Create a symbolic link.
    e.g
        j.sals.fs.symlink(src="/home/rafy/testing_text.txt",dst="/home/rafy/link_test")

    Args:
        src (str): Source of link
        dst (str): Destination path of link
        overwrite (bool, optional): If link exists will delete it. Defaults to False.
    """
    if overwrite and exists(dst):
        os.unlink(dst)

    os.symlink(src, dst)


copy_tree = dir_util.copy_tree
chdir = os.chdir


def change_dir(path: str) -> str:
    """Change current working directory to `path`
    e.g
         j.sals.fs.change_dir("/home/rafy/Documents")  -> '/home/rafy/Documents'

    Args:
        path (str): path to switch current working directory to

    Returns:
        str: new current working dir
    """
    os.chdir(path)
    return path


def chmod(path: str, mode):
    """change file mode for path to mode
    e.g
        j.sals.fs.chmod("/home/rafy/testing_dir",777)

    Args:
        path (str): path
        mode (int): file mode

    """
    return pathlib.Path(path).chmod(mode)


def lchmod(path: str, mode):
    """change file mode for path to mode (handles links too)

    Args:
        path (str): path
        mode (int): file mode

    """
    return pathlib.Path(path).lchmod(mode)


def stat(path: str):
    """Gets stat of path `path`
    e.g
        j.sals.fs.stat("/home/rafy/test_dir/test")  -> os.stat_result(st_mode=33204, st_ino=795348, st_dev=2049, st_nlink=1, st_uid=1000, st_gid=1000, st_size=0, st_atime=1586445434, st_mtime=1586445434, st_ctime=1586445434)
    Args:
        path (str): path to get its stat

    Returns:
        stat_result: returns stat struct.
    """

    return pathlib.Path(path).stat()


def lstat(path: str):
    """Gets stat of path `path` (handles links)
    e.g
         j.sals.fs.lstat("/home/rafy/testing_link")  -> os.stat_result(st_mode=41471, st_ino=7081257, st_dev=2049, st_nlink=1, st_uid=1000, st_gid=1000, st_size=16, st_atime=1586445737, st_mtime=1586445734, st_ctime=1586445734)

    Args:
        path (str): path to get its stat

    Returns:
        stat_result: returns stat struct.
    """

    return pathlib.Path(path).lstat()


def resolve(path: str) -> str:
    """resolve `.` and `..` in path
    e.g
        j.sals.fs.resolve("")  -> PosixPath('/home/rafy/Documents')
        j.sals.fs.resolve("./testing_text.txt")  -> PosixPath('/home/rafy/Documents/testing_text.txt')

    Args:
        path (str): path with optionally `.` and `..`

    Returns:
        str: resolved path
    """
    return str(pathlib.Path(path).resolve())


def extension(path: str, include_dot=True):
    """Gets the extension of path
    e.g
        '/home/ahmed/myfile.py' -> `.py` if include_dot else `py`

    Args:
        path (str): [description]
        include_dot (bool, optional): controls whether to include the dot or not. Defaults to True.

    Returns:
        str: extension
    """
    splitted = os.path.splitext(path)
    ext = ""
    if len(splitted) == 1:
        return ext

    if include_dot:
        return splitted[1]
    else:
        return splitted[1].strip(".")


ext = extension


def chown():
    raise NotImplementedError()


def read_link(path):
    raise NotImplementedError()


def remove_links(path):
    raise NotImplementedError()


def change_filenames(from_, to, where):
    pass


def replace_words_in_files(from_, to, where):
    pass


move = shutil.move


def default_filter_fun(entry):
    return True


def walk(path: str, pat="*", filter_fun=default_filter_fun):
    """walk recursively on path
    e.g
        for el in walk('/tmp', filter_fun=j.sals.fs.is_file) : ..
        for el in walk('/tmp', filter_fun=j.sals.fs.is_dir) : ..
        for el in walk('/tmp', filter_fun= lambda x: len(x)>4 and (j.sals.fs.is_file(x) or j.sals.fs.is_dir(x)) ) : ..


    Args:
        path (str): path to walk over
        pat (str, optional): pattern to match against. Defaults to "*".
        filter_fun (Function, optional): filtering function. Defaults to default_filter_fun which accepts anything.
    """
    p = pathlib.Path(path)
    for entry in p.rglob(pat):
        # use rglob instead of glob("**/*")
        if filter_fun(entry):
            yield str(entry)


def walk_non_recursive(path: str, filter_fun=default_filter_fun):
    """walks non recursively on path
    e.g
        for el in walk('/tmp', filter=j.sals.fs.is_file) : ..
        for el in walk('/tmp', filter=j.sals.fs.is_dir) : ..
        for el in walk('/tmp', filter= lambda x: len(x)>4 and (j.sals.fs.is_file(x) or j.sals.fs.is_dir(x)) ) : ..


    Args:
        path (str): path to walk over
        pat (str, optional): pattern to match against. Defaults to "*".
        filter_fun (Function, optional): filtering function. Defaults to default_filter_fun which accepts anything.
    """
    p = pathlib.Path(path)
    for entry in p.iterdir():
        if filter_fun(entry):
            yield str(entry)


def walk_files(path: str, recursive=True):
    """
    walk over files in path and applies function `fun`
    e.g

        for el in walk_files('/tmp') : ..

    Args:
        path (str): path to walk over
        recursive (bool, optional): recursive or not. Defaults to True.


    """

    if recursive:
        return walk(path, filter_fun=is_file)
    else:
        return walk_non_recursive(path, filter_fun=is_file)


def walk_dirs(path, recursive=True):
    """
        walk over directories in path and applies function `fun`
    e.g

        for el in walk_dirs('/tmp') : ..


    Args:
        path (str): path to walk over
        recursive (bool, optional): recursive or not. Defaults to True.


    """
    if recursive:
        return walk(path, filter_fun=is_dir)
    else:
        return walk_non_recursive(path, filter_fun=is_dir)


def fs_check(**arguments):
    """Abstracts common checks over your file system related functions.
    To reduce the boilerplate of expanding paths, checking for existence or ensuring non empty values.

    Checks are defined for each argument separately in a form of a set
        e.g
        @fs_check(path={'required', 'exists', 'expand'})
        @fs_check(path1={'required', 'exists', 'expand'}, path2={'required', 'exists', 'expand'})

    Available checks:
        - `required`: ensures the argument is passed with a non-empty value.
        - `expand`  : expands the tilde `~` in path.
        - `exists`  : the path exists.
        - `file`    : the path is a file.
        - `dir`     : the path is a dir.

    """

    for argument, validators in arguments.items():
        if not isinstance(validators, set):
            raise ValueError(f"Expected tuple of validators for argument {argument}")
        for validator in validators:
            if validator not in {"required", "exists", "file", "dir", "expand"}:
                raise ValueError(f"Unsupported validator '{validator}' for argument {argument}")

    def decorator(func):
        import inspect

        signature = inspect.signature(func)
        for argument in arguments:
            if signature.parameters.get(argument) is None:
                raise j.exceptions.Value(f"Argument {argument} not found in function declaration of {func.__name__}")

        def wrapper(*args, **kwargs):
            args = list(args)
            position = 0
            for parameter in signature.parameters.values():
                if parameter.name in arguments:
                    value = args[position] if position < len(args) else kwargs[parameter.name]
                    if isinstance(value, str):
                        value = expanduser(expandvars(value))
                    if position < len(args):
                        args[position] = value
                    else:
                        kwargs[parameter.name] = value

                    validators = arguments[parameter.name]
                    if value and validators.intersection({"exists", "file", "dir"}) and not exists(value):
                        msg = f"Argument {parameter.name} in {func.__name__} expects an existing path value! {value} does not exist."
                        raise j.exceptions.Value(msg)

                    if "required" in validators and (value is None or value.strip() == ""):
                        raise j.exceptions.Value(
                            f"Argument {parameter.name} in {func.__name__}  should not be None or empty string!"
                        )

                    if "required" in validators:
                        value = norm_path(value)
                        if position < len(args):
                            args[position] = value
                        else:
                            kwargs[parameter.name] = value

                    if value and validators.intersection({"file"}) and not isfile(value):
                        raise j.exceptions.Value(
                            f"Argument {parameter.name} in {func.__name__} expects a file path! {value} is not a file."
                        )
                    if value and validators.intersection({"dir"}) and not isdir(value):
                        raise j.exceptions.Value(
                            f"Argument {parameter.name} in {func.__name__} expects a directory path! {value} is not a directory."
                        )
                position += 1
            return fun(*args, **kwargs)

        return wrapper

    return decorator



import os

from . import ConfigNotFound, EncryptedConfigStore
from .serializers import JsonSerializer

from jumpscale.sals.fs import exists, make_path, read_file_binary, rmtree, write_file_binary


class FileSystemStore(EncryptedConfigStore):
    """
    Filesystem store is an EncryptedConfigStore

    It saves the config relative to `config_env.get_store_config("filesystem")`

    To store every instance config in a different path, it uses the given `Location`.
    """

    def __init__(self, location):
        """
        create a new `FileSystemStore` that stores config at the given location under configured root.

        The root directory can be configured, see `jumpscale.core.config`

        Args:
            location (Location): where config will be stored per instance
        """
        super(FileSystemStore, self).__init__(location, JsonSerializer())
        self.root = self.config_env.get_store_config("filesystem")["path"]

    @property
    def config_root(self):
        """
        get the root directory where all configurations are written

        Returns:
            str: path
        """
        return os.path.join(self.root, self.location.path)

    def get_instance_root(self, instance_name):
        """
        get the directory where instance config is written

        Args:
            instance_name (str): name

        Returns:
            str: path
        """
        return os.path.join(self.config_root, instance_name)

    def get_path(self, instance_name):
        """
        get the path to data file where instance config is written

        Args:
            instance_name (str): name

        Returns:
            str: path
        """
        return os.path.join(self.get_instance_root(instance_name), "data")

    def read(self, instance_name):
        """
        read config data from the data file

        Args:
            instance_name (str): name

        Returns:
            str: data
        """
        path = self.get_path(instance_name)
        if not exists(path):
            raise ConfigNotFound(f"cannot find config for {instance_name} at {path}")
        return read_file_binary(path)

    def list_all(self):
        """
        list all instance names (directories under config root)

        Returns:
            list: instance/directory names
        """
        if not os.path.exists(self.config_root):
            return []
        return os.listdir(self.config_root)

    def write(self, instance_name, data):
        """
        write config data to data file

        Args:
            instance_name (str): config
            data (str): data

        Returns:
            bool: written or not
        """
        path = self.get_path(instance_name)
        make_path(path)
        return write_file_binary(path, data.encode())

    def delete(self, instance_name):
        """
        delete instance config directory

        Args:
            instance_name (str):
        """
        path = self.get_instance_root(instance_name)
        if os.path.exists(path):
            rmtree(path)



"""
This module is for event handling, where any component can listen to certain event notifications

The modules is based on https://github.com/xmonader/events.

Events can be of any type (class), there are two ways for listening to an event:

Using decorators:

```python
from jumpscale.core.base import events

class ImportantEvent:
    def __init__(self):
        self.name = "custom event"


@events.handle(ImportantEvent)
def handle_event(ev):
    print(ev.name)
```

Or using it in your classes, just inherit from `events.Handler` and implement `handle` method.

```python3
class Impl(events.Handler):
    def __init__(self):
        events.add_listeners(self, ImprotantEvent)

    def handle(self, ev):
        print(ev.name)
```


For more control, have a look at:

- `events.add_global_listener`: for adding a global listener
- `events.handle_any` and `events.handle_many`

This can be used with base classes too, you just need to define your own custom events.
For an example, see `redis.RedisClient`
"""
from collections import defaultdict
from functools import wraps


class Any:
    pass


class Handler:
    def handle(self, ev):
        pass


listeners = defaultdict(list)
listeners[Any] = []


def add_listenter(handler, *event_types):
    if not event_types:
        raise ValueError("must specify at least 1 event type/class")
    for event_type in event_types:
        if handler not in listeners[event_type]:
            listeners[event_type].append(handler)


def add_global_listener(handler):
    add_listenter(handler, Any)


def handle_many(*event_types):
    def decorator(fun):
        add_listenter(fun, *event_types)

        @wraps(fun)
        def wrapper(*args, **kwargs):
            fun(*args, **kwargs)

        return wrapper

    return decorator


def handle_any(fun):
    add_global_listener(fun)

    @wraps(fun)
    def wrapper(*args, **kwargs):
        fun(*args, **kwargs)

    return wrapper


def handle(event_type):
    def decorator(fun):
        add_listenter(fun, event_type)

        @wraps(fun)
        def wrapper(*args, **kwargs):
            fun(*args, **kwargs)

        return wrapper

    return decorator


def notify(event):
    event_type = event.__class__
    interested = listeners.get(event_type, []) + listeners[Any]

    for handler in interested:
        if isinstance(handler, Handler):
            handler.handle(event)
        else:
            handler(event)



import binascii
import nacl.utils

from nacl.public import PrivateKey, PublicKey, Box
from nacl.secret import SecretBox
from nacl.signing import SigningKey, VerifyKey
from nacl.exceptions import BadSignatureError


class NACL:
    KEY_SIZE = 32

    def __init__(self, private_key=None, symmetric_key=None):
        """Constructor for nacl object

        Args:
            private_key (bytes, optional): The private key used to sign and encrypt the data. Generated randomly if not given.
            symmetric_key (bytes, optional): The key used for symmetric encryption. Generated randomly if not given.
        """
        self.signing_key = SigningKey(private_key) if private_key else SigningKey.generate()
        self.verify_key = self.signing_key.verify_key
        self.private_key = self.signing_key.to_curve25519_private_key()
        self.public_key = self.private_key.public_key
        self.symmetric_key = nacl.utils.random(NACL.KEY_SIZE) if symmetric_key is None else symmetric_key
        self.symmetric_box = SecretBox(self.symmetric_key)

    def encrypt(self, message, reciever_public_key):
        """Encrypt the message to send to a receiver. (public key encryption)

        Args:
            message (bytes): The message to be encrypted.
            reciever_public_key (bytes): The receiver's public key.

        Returns:
            bytes: The encrypted message
        """
        return Box(self.private_key, PublicKey(reciever_public_key)).encrypt(message)

    def decrypt(self, message, sender_public_key):
        """Decrypt a received message. (public key encryption)

        Args:
            message (bytes): The encrypted message.
            sender_public_key (bytes): The public key of the sender.

        Returns:
            bytes: The decrypted message
        """
        return Box(self.private_key, PublicKey(sender_public_key)).decrypt(message)

    def encrypt_symmetric(self, message):
        """Encrypt the message to send to a receiver. (secret key encryption)

        Args:
            message (bytes): The message to be encrypted.

        Returns:
            bytes: The encrypted message
        """
        return self.symmetric_box.encrypt(message)

    def decrypt_symmetric(self, message, symmetric_key):
        """Decrypt the receiver message. (secret key encryption)

        Args:
            message (bytes): The message to be decrypted.

        Returns:
            bytes: The decrypted message
        """
        return SecretBox(symmetric_key).decrypt(message)

    def sign(self, message):
        """Sign the message and return the messsage and the signature.

        Args:
            message (bytes): The message to be signed

        Returns:
            bytes, bytes: The message and the signature.
        """
        signed = self.signing_key.sign(message)
        return message, signed.signature

    def sign_hex(self, message):
        """Sign the message and return the messsage and the signature.

        Args:
            message (bytes): The message to be signed

        Returns:
            bytes: The message and the signature.
        """
        signed = self.signing_key.sign(message)
        signedhex = binascii.hexlify(signed.signature)
        return signedhex

    def verify(self, message, signature, verification_key):
        """Verify that the signature using the verification key

        Args:
            message (bytes): The received message.
            signature (bytes): The recieved signature.
            verification_key (bytes): The verification key.

        Returns:
            bool: True if the verification succeeds.
        """
        try:
            VerifyKey(verification_key).verify(message, signature)
            return True
        except BadSignatureError:
            return False

    def get_signing_seed(self):
        """Returns the signing seed (same as the private key).

        Returns:
            bytes: The 32-bit signing key.
        """
        return bytes(self.signing_key._seed)

    def get_verification_key(self):
        """Returns the verification key.

        Returns:
            bytes: The verification key.
        """
        return bytes(self.verify_key)

    def get_verify_key_hex(self):
        return binascii.hexlify(self.verify_key.encode()).decode()

    def get_public_key_hex(self):
        return binascii.hexlify(self.public_key.encode()).decode()

    def get_public_key(self):
        """Getter for the public key.

        Returns:
            bytes: The public key.
        """
        return bytes(self.public_key)

    def get_private_key(self):
        """Getter for the private key.

        Returns:
            bytes: The private key.
        """
        return bytes(self.private_key)

    def get_symmetric_key(self):
        """Getter for the symmetric key.

        Returns:
            bytes: The symmetric key.
        """
        return bytes(self.symmetric_key)



import inspect
import json
import sys
import os
from redis import Redis
from enum import Enum
from functools import partial
from io import BytesIO
from signal import SIGKILL, SIGTERM
import json
import better_exceptions
import gevent
from gevent.pool import Pool
from gevent import time
from gevent.server import StreamServer
from jumpscale.core.base import Base, fields
from jumpscale.loader import j
from redis.connection import DefaultParser, Encoder
from redis.exceptions import ConnectionError
from .baseactor import BaseActor
from .systemactor import CoreActor, SystemActor


def serialize(obj):
    if not isinstance(obj, (str, int, float, list, tuple, dict, bool)):
        module = inspect.getmodule(obj).__file__[:-3]
        return dict(__serialized__=True, module=module, type=obj.__class__.__name__, data=obj.to_dict())
    return obj


def deserialize(obj):
    if isinstance(obj, dict) and obj.get("__serialized__"):
        module = sys.modules[obj["module"]]
        object_instance = getattr(module, obj["type"])()
        object_instance.from_dict(obj["data"])
        return object_instance
    return obj


class GedisErrorTypes(Enum):
    NOT_FOUND = 0
    BAD_REQUEST = 1
    ACTOR_ERROR = 3
    INTERNAL_SERVER_ERROR = 4


class RedisConnectionAdapter:
    def __init__(self, sock):
        self.socket = sock
        self._sock = sock
        self.socket_timeout = 600
        self.socket_connect_timeout = 600
        self.socket_keepalive = True
        self.retry_on_timeout = True
        self.socket_keepalive_options = {}
        self.encoder = Encoder("utf", "strict", False)


class ResponseEncoder:
    def __init__(self, socket):
        self.socket = socket
        self.buffer = BytesIO()

    def encode(self, value):
        """Respond with data."""
        if value is None:
            self._write_buffer("$-1\r\n")
        elif isinstance(value, int):
            self._write_buffer(":{}\r\n".format(value))
        elif isinstance(value, bool):
            self._write_buffer(":{}\r\n".format(1 if value else 0))
        elif isinstance(value, str):
            if "\n" in value:
                self._bulk(value)
            else:
                self._write_buffer("+{}\r\n".format(value))
        elif isinstance(value, bytes):
            self._bulkbytes(value)
        elif isinstance(value, list):
            if value and value[0] == "*REDIS*":
                value = value[1:]
            self._array(value)
        elif hasattr(value, "__repr__"):
            self._bulk(value.__repr__())
        else:
            value = j.data.serializers.json.dumps(value, encoding="utf-8")
            self.encode(value)

        self._send()

    def status(self, msg="OK"):
        """Send a status."""
        self._write_buffer("+{}\r\n".format(msg))
        self._send()

    def error(self, msg):
        """Send an error."""
        # print("###:%s" % msg)
        self._write_buffer("-ERR {}\r\n".format(msg))
        self._send()

    def _bulk(self, value):
        """Send part of a multiline reply."""
        data = ["$", str(len(value)), "\r\n", value, "\r\n"]
        self._write_buffer("".join(data))

    def _array(self, value):
        """send an array."""
        self._write_buffer("*{}\r\n".format(len(value)))
        for item in value:
            self.encode(item)

    def _bulkbytes(self, value):
        data = [b"$", str(len(value)).encode(), b"\r\n", value, b"\r\n"]
        self._write_buffer(b"".join(data))

    def _write_buffer(self, data):
        if isinstance(data, str):
            data = data.encode()

        self.buffer.write(data)

    def _send(self):
        self.socket.sendall(self.buffer.getvalue())
        self.buffer = BytesIO()  # seems faster then truncating


SERIALIZABLE_TYPES = (str, int, float, list, tuple, dict, bool)
RESERVED_ACTOR_NAMES = ("core", "system")


class GedisServer(Base):
    host = fields.String(default="127.0.0.1")
    port = fields.Integer(default=16000)
    enable_system_actor = fields.Boolean(default=True)
    run_async = fields.Boolean(default=True)
    _actors = fields.Typed(dict, default={})

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._core_actor = CoreActor()
        self._system_actor = SystemActor()
        self._loaded_actors = {"core": self._core_actor}

    @property
    def actors(self):
        """Lists saved actors

        Returns:
            list -- List of saved actors
        """
        return self._actors

    def actor_add(self, actor_name: str, actor_path: str):
        """Adds an actor to the server

        Arguments:
            actor_name {str} -- Actor name
            actor_path {str} -- Actor absolute path

        Raises:
            j.exceptions.Value: raises if actor name is matched one of the reserved actor names
            j.exceptions.Value: raises if actor name is not a valid identifier
        """
        if actor_name in RESERVED_ACTOR_NAMES:
            raise j.exceptions.Value("Invalid actor name")

        if not actor_name.isidentifier():
            raise j.exceptions.Value(f"Actor name should be a valid identifier")

        self._actors[actor_name] = actor_path

    def actor_delete(self, actor_name: str):
        """Removes an actor from the server

        Arguments:
            actor_name {str} -- Actor name
        """
        self._actors.pop(actor_name, None)

    def start(self):
        """Starts the server
        """
        j.application.start("gedis")

        # handle signals
        for signal_type in (SIGTERM, SIGKILL):
            gevent.signal(signal_type, self.stop)

        # register system actor if enabled
        if self.enable_system_actor:
            self._register_actor("system", self._system_actor)

        self._core_actor.set_server(self)
        self._system_actor.set_server(self)

        # register saved actors
        for actor_name, actor_path in self._actors.items():
            self._system_actor.register_actor(actor_name, actor_path)

        # start the server
        self._server = StreamServer((self.host, self.port), self._on_connection, spawn=Pool())
        self._server.reuse_addr = True
        self._server.start()

    def stop(self):
        """Stops the server
        """
        j.logger.info("Shutting down...")
        self._server.stop()

    def _register_actor(self, actor_name: str, actor_module: BaseActor):
        self._loaded_actors[actor_name] = actor_module

    def _unregister_actor(self, actor_name: str):
        self._loaded_actors.pop(actor_name, None)

    def _execute(self, method, args, kwargs):
        response = {}
        try:
            response["result"] = method(*args, **kwargs)

        except TypeError as e:
            response["error"] = str(e)
            response["error_type"] = GedisErrorTypes.BAD_REQUEST.value

        except:
            ttype, tvalue, tb = sys.exc_info()
            response["error"] = better_exceptions.format_exception(ttype, tvalue, tb)
            response["error_type"] = GedisErrorTypes.ACTOR_ERROR.value

        return response

    def _on_connection(self, socket, address):
        j.logger.info("New connection from {}", address)
        parser = DefaultParser(65536)
        connection = RedisConnectionAdapter(socket)
        try:
            encoder = ResponseEncoder(socket)
            parser.on_connect(connection)

            while True:
                response = dict(
                    success=True, result=None, error=None, error_type=None, is_async=False, task_id=None
                )
                try:   
                    request = parser.read_response()

                    if len(request) < 2:
                        response["error"] = "invalid request"
                        response["error_type"] = GedisErrorTypes.BAD_REQUEST.value

                    else:
                        actor_name = request.pop(0).decode()
                        method_name = request.pop(0).decode()
                        actor_object = self._loaded_actors.get(actor_name)

                        if not actor_object:
                            response["error"] = "actor not found"
                            response["error_type"] = GedisErrorTypes.NOT_FOUND.value

                        elif not hasattr(actor_object, method_name):
                            response["error"] = "method not found"
                            response["error_type"] = GedisErrorTypes.NOT_FOUND.value

                        else:
                            j.logger.info(
                                "Executing method {} from actor {} to client {}", method_name, actor_name, address
                            )

                            if request:
                                args, kwargs = json.loads(request.pop(0), object_hook=deserialize)
                            else:
                                args, kwargs = (), {}

                            method = getattr(actor_object, method_name)
                            result = self._execute(method, args, kwargs)
                            response.update(result)

                except ConnectionError:
                    j.logger.info("Client {} closed the connection", address)

                except Exception as exception:
                    j.logger.exception("internal error", exception=exception)
                    response["error"] = "internal server error"
                    response["error_type"] = GedisErrorTypes.INTERNAL_SERVER_ERROR.value

                response["success"] = response["error"] is None
                encoder.encode(json.dumps(response, default=serialize))

            parser.on_disconnect()

        except BrokenPipeError:
            pass



from .model_model import ModelBase
import os
import importlib

def add_model(file_name):
    m = importlib.import_module("." + file_name[:-3], "jumpscale.data.bcdb.models")
    for attr in dir(m):
        pyattr = getattr(m, attr)
        if isinstance(pyattr, type) and issubclass(pyattr, ModelBase):
            globals()[attr] = pyattr

files = os.listdir(os.path.dirname(__file__))
for f in files:
    if f.endswith("_model.py"):
        add_model(f)


"""tmux module allows manipulation of tmux sessions, pane and launching background commands in it

"""
import libtmux
from jumpscale.core.logging import export_module_as as logger

from .command_builder import cmd_from_args

__all__ = ["execute_in_window"]

server = libtmux.Server()
JS_SESSION_NAME = "js-ng"


def create_session(session_name, kill_if_exists=False):
    return server.new_session(session_name, kill_session=kill_if_exists)


def kill_session(session_name):
    return server.kill_session(session_name)


def get_js_session():
    try:
        s = server.find_where({"session_name": JS_SESSION_NAME})
        if not s:
            return create_session(JS_SESSION_NAME)
    except libtmux.exc.LibTmuxException:
        return create_session(JS_SESSION_NAME)
    return s


def get_js_window(window_name):
    s = get_js_session()
    w = s.find_where({"window_name": window_name})
    if not w:
        w = s.new_window(window_name)
        w.rename_window(window_name)

    return w


def get_window(window_name):
    w = server.find_where({"window_name": window_name})
    if not w:
        w = get_js_window(window_name)
    return w


@cmd_from_args
def execute_in_window(cmd, window_name):
    """
    execute a command in a new tmux window

    Args:
        cmd (str or list): command as a string or an argument list, e.g. `"ls -la"` or `["ls", "la"]`
        window_name (str): window name
    """
    try:
        server.list_sessions()
    except:
        logger().error("tmux isn't running")
        server.new_session(JS_SESSION_NAME)
    w = get_window(window_name)
    w.attached_pane.send_keys(cmd)



import inspect
import json
import os
import sys
from functools import partial

from jumpscale.clients.base import Client
from jumpscale.core.base import fields
from jumpscale.loader import j
from jumpscale.servers.gedis.server import GedisErrorTypes, deserialize, serialize
from jumpscale.tools.codeloader import load_python_module


class ActorResult:
    def __init__(self, **kwargs):
        self.success = kwargs.get("success", True)
        self.result = kwargs.get("result", None)
        self.error = kwargs.get("error", None)
        self.error_type = kwargs.get("error_type", None)
        self.is_async = kwargs.get("is_async", False)
        self.task_id = kwargs.get("task_id", None)

    def __dir__(self):
        return list(self.__dict__.keys())

    def __repr__(self):
        return str(self.__dict__)


class ActorProxy:
    def __init__(self, actor_name, actor_info, client):
        """ActorProxy to remote actor on the server side

        Arguments:
            actor_name {str} -- [description]
            actor_info {dict} -- actor information dict e.g { method_name: { args: [], 'doc':...} }
            gedis_client {GedisClient} -- gedis client reference
        """
        self.actor_name = actor_name
        self.actor_info = actor_info
        self.client = client

    def __dir__(self):
        """Delegate the available functions on the ActorProxy to `actor_info` keys

        Returns:
            list -- methods available on the ActorProxy
        """
        return list(self.actor_info["methods"].keys())

    def __getattr__(self, method):
        """Return a function representing the remote function on the actual actor

        Arguments:
            attr {str} -- method name

        Returns:
            function -- function waiting on the arguments
        """

        def function(*args, **kwargs):
            return self.client.execute(self.actor_name, method, *args, **kwargs)

        func = partial(function)
        func.__doc__ = self.actor_info["methods"][method]["doc"]
        return func


class ActorsCollection:
    def __init__(self, actors):
        self._actors = actors

    def __dir__(self):
        return list(self._actors.keys())

    def __getattr__(self, actor_name):
        if actor_name in self._actors:
            return self._actors[actor_name]


class GedisClient(Client):
    name = fields.String(default="local")
    hostname = fields.String(default="localhost")
    port = fields.Integer(default=16000)
    raise_on_error = fields.Boolean(default=False)
    disable_deserialization = fields.Boolean(default=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._redisclient = None
        self._loaded_actors = {}
        self._loaded_modules = []
        self.actors = None
        self._load_actors()

    @property
    def redis_client(self):
        if self._redisclient is None:
            self._redisclient = j.clients.redis.get(name=f"gedis_{self.name}", hostname=self.hostname, port=self.port)
        return self._redisclient

    def _load_module(self, path):
        if path not in self._loaded_modules:
            load_python_module(path)
            self._loaded_modules.append(path)

    def _load_actors(self):
        self._loaded_actors = {}
        for actor_name in self.list_actors():
            actor_info = self._get_actor_info(actor_name)
            self._load_module(actor_info["path"])
            self._loaded_actors[actor_name] = ActorProxy(actor_name, actor_info, self)

        self.actors = ActorsCollection(self._loaded_actors)

    def _get_actor_info(self, actor_name):
        return self.execute(actor_name, "info", die=True).result

    def list_actors(self) -> list:
        """List actors

        Returns:
            list -- List of loaded actors
        """
        return self.execute("core", "list_actors", die=True).result

    def reload(self):
        """Reload actors
        """
        self._load_actors()

    def execute(self, actor_name: str, actor_method: str, *args, die: bool = False, **kwargs) -> ActorResult:
        """Execute actor's method

        Arguments:
            actor_name {str} -- actor name
            actor_method {str} -- actor method

        Keyword Arguments:
            die {bool} --  flag to raise an error when request fails (default: {False})

        Raises:
            RemoteException: Raises if the request failed and raise_on_error flag is set

        Returns:
            ActorResult -- request result
        """
        payload = json.dumps((args, kwargs), default=serialize)
        response = self.redis_client.execute_command(actor_name, actor_method, payload)

        deserializer = deserialize if not self.disable_deserialization else None
        response = json.loads(response, object_hook=deserializer)

        if not response["success"]:
            if die or self.raise_on_error:
                raise RemoteException(response["error"])

            response["error_type"] = GedisErrorTypes(response["error_type"])

        return ActorResult(**response)


class RemoteException(Exception):
    pass



import os


from jumpscale.core.config import get_current_version
from jumpscale.core.exceptions import Value
from jumpscale.clients.docker.docker import DockerClient

from . import settings
from .container import Container
from .identitymanager import IdentityManager

DEFAULT_CONTAINER_NAME = "3bot-ng"
DEFAULT_IMAGE = "threefoldtech/js-ng"
PERSISTENT_STORE = os.path.expanduser("~/.config/jumpscale/containers")

docker_client = DockerClient()


class ThreeBot(Container):
    """
    Manage your threebot
    """

    @staticmethod
    def install(
        name=None, image=None, identity=None, email=None, words=None, explorer=None, development: bool = None,
    ):
        """Creates a threebot container

        Args:
            name (str, optional): name of the container. Defaults to 3bot-ng
            image (str, optional): container image. Defaults to "threefoldtech/js-ng:latest".
            identity (str, optional): threebot name. Defaults to None.
            email (str, optional): threebot email. Defaults to None.
            words (str, optional): seed phrase of the user. Defaults to None.
            explorer (str, optional): which explorer network to use: mainnet, testnet, devnet. Defaults to None.
            development (bool, optional): if true will mount codedir. Defaults to False.

        Raises:
            Value: Container with specified name already exists
            Value: explorer not in mainnet, testnet, devnet
        """
        if development is None:
            development = settings.expert
        name = name or DEFAULT_CONTAINER_NAME
        current_version = get_current_version()
        image = image or f"{DEFAULT_IMAGE}:{current_version}"

        pers_path = f"{PERSISTENT_STORE}/{name}"
        configure = not os.path.exists(pers_path)
        if configure:
            identity = IdentityManager(identity, email, words, explorer)
            identity, email, words, explorer = identity.ask_identity()

        os.makedirs(PERSISTENT_STORE, exist_ok=True)
        volumes = {pers_path: {"bind": "/root/.config/jumpscale", "mode": "rw"}}

        container = Container.install(name, image, development, volumes)
        container.exec_run(["redis-server", "--daemonize yes"])

        if configure:
            container.exec_run(["jsng", f"j.core.identity.new('default', '{identity}', '{email}', '{words}')"])
            container.exec_run(["jsng", "j.core.identity.set_default('default')"])

    @staticmethod
    def jsng(name=DEFAULT_CONTAINER_NAME):
        """Get's shell in threebot

        Args:
            name (str): name of the container (default: 3bot-ng)
        """
        Container.exec(name, "jsng")

    @staticmethod
    def shell(name=DEFAULT_CONTAINER_NAME):
        """Get's shell in threebot

        Args:
            name (str): name of the container (default: 3bot-ng)
        """
        Container.exec(name, "bash")

    @staticmethod
    def start(name=DEFAULT_CONTAINER_NAME):
        """Start threebot container with threebot server

        Args:
            name (str): name of the container (default: 3bot-ng)
        """
        Container.start(name)
        Container.exec(name, ["threebot", "start", "--background"])

    @staticmethod
    def stop(name=DEFAULT_CONTAINER_NAME):
        """Stop threebot installation with container

        Args:
            name (str): name of the container (default: 3bot-ng)
        """
        if name in docker_client.list():
            Container.exec(name, ["threebot", "stop"])
            Container.stop(name)
        else:
            print("Container is already stopped")

    @staticmethod
    def restart(name=DEFAULT_CONTAINER_NAME):
        """restart threebot installation with container

        Args:
            name (str): name of the container (default: 3bot-ng)
        """
        ThreeBot.stop(name=name)
        ThreeBot.start(name=name)
